from decimal import Decimal

from django.conf import settings
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy
from otree.currency import Currency, RealWorldCurrency
from django.forms.widgets import * # noqa
from django import forms


from otree.currency.locale import CURRENCY_SYMBOLS


class _BaseMoneyInput(forms.NumberInput):
    # step = 0.01
    template_name = 'otree/forms/moneyinput.html'

    def get_context(self, *args, **kwargs):
        context = super(_BaseMoneyInput, self).get_context(*args, **kwargs)
        context['currency_symbol'] = self.CURRENCY_SYMBOL
        return context

    def _format_value(self, value):
        if isinstance(value, (Currency, RealWorldCurrency)):
            value = Decimal(value)
        return force_text(value)


class _RealWorldCurrencyInput(_BaseMoneyInput):
    '''it's a class attribute so take care with patching it in tests'''
    CURRENCY_SYMBOL = CURRENCY_SYMBOLS.get(
        settings.REAL_WORLD_CURRENCY_CODE,
        settings.REAL_WORLD_CURRENCY_CODE,
    )


class _CurrencyInput(_RealWorldCurrencyInput):
    '''it's a class attribute so take care with patching it in tests'''
    if settings.USE_POINTS:
        if hasattr(settings, 'POINTS_CUSTOM_NAME'):
            CURRENCY_SYMBOL = settings.POINTS_CUSTOM_NAME
        else:
            # Translators: the label next to a "points" input field
            CURRENCY_SYMBOL = ugettext_lazy('points')


class RadioSelectHorizontal(forms.RadioSelect):
    template_name = 'otree/forms/radio_select_horizontal.html'


class Slider(forms.NumberInput):
    input_type = 'range'
    template_name = 'otree/forms/slider.html'
    show_value = True

    def __init__(self, *args, **kwargs):
        try:
            # fix bug where currency "step" values were ignored.
            step = kwargs['attrs']['step']
            kwargs['attrs']['step'] = self._format_value(step)
        except KeyError:
            pass
        show_value = kwargs.pop('show_value', None)
        if show_value is not None:
            self.show_value = show_value
        super().__init__(*args, **kwargs)

    def _format_value(self, value):
        if isinstance(value, (Currency, RealWorldCurrency)):
            value = Decimal(value)
        return force_text(value)

    def get_context(self, *args, **kwargs):
        context = super().get_context(*args, **kwargs)
        context['show_value'] = self.show_value
        return context

class SliderInput(Slider):
    '''old name for Slider widget'''
    pass



'''
call this file staticfiles instead of static because PyCharm is confused
with otree/static (even though that's not a module)
'''

from django.templatetags.static import StaticNode as DjStaticNode
from django.template.base import FilterExpression


class BackslashError(ValueError):
    pass


class StaticNode(DjStaticNode):
    def __init__(self, varname=None, path:FilterExpression=None):
        # path.token is the literal string, not the value of the variable
        # it resolves to,
        # so there should never be a \
        if path and ('\\' in path.token):
            msg = (
                r'{{% static {} %}} '
                r'contains a backslash ("\"); '
                r'you should change it to a forward slash ("/").'
            ).format(path.token)
            raise BackslashError(msg)
        # should we also handle cases where the \ was in a variable?
        # then have to consider about os.path.join which will use os.sep
        super().__init__(varname, path)



from django.http import HttpResponseServerError
import time
from otree.common_internal import missing_db_tables
import logging

logger = logging.getLogger('otree.perf')

def perf_middleware(get_response):
    # One-time configuration and initialization.

    def middleware(request):
        start = time.time()

        response = get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        # heroku has 'X-Request-ID', which Django translates to
        # the following:
        request_id = request.META.get('HTTP_X_REQUEST_ID')
        if request_id:
            # only log this info on Heroku
            elapsed = time.time() - start
            msec = int(elapsed * 1000)
            msg = f'own_time={msec}ms request_id={request_id}'
            logger.info(msg)

        return response

    return middleware


class CheckDBMiddleware:
    synced = None

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not CheckDBMiddleware.synced:
            # very fast, 0.01-0.02 seconds for the whole check
            missing_tables = missing_db_tables()
            if missing_tables:
                listed_tables = missing_tables[:3]
                unlisted_tables = missing_tables[3:]
                msg = (
                    "Your database is not ready. Try resetting the database "
                    "(Missing tables for {}, and {} other models). "
                ).format(
                    ', '.join(listed_tables), len(unlisted_tables))
                return HttpResponseServerError(msg)
            else:
                CheckDBMiddleware.synced = True
        return self.get_response(request)



import time

import vanilla
from django.urls import reverse
from django.http import HttpResponseRedirect, JsonResponse
from otree.channels import utils as channel_utils
from otree.models_concrete import ParticipantRoomVisit
from otree.room import ROOM_DICT
from otree.views.admin import CreateSession


class Rooms(vanilla.TemplateView):
    template_name = 'otree/admin/Rooms.html'

    url_pattern = r"^rooms/$"

    def get_context_data(self, **kwargs):
        return {'rooms': ROOM_DICT.values()}


class RoomWithoutSession(CreateSession):
    template_name = 'otree/admin/RoomWithoutSession.html'
    room = None

    url_pattern = r"^room_without_session/(?P<room_name>.+)/$"

    def dispatch(self, request, *args, **kwargs):
        self.room = ROOM_DICT[kwargs['room_name']]
        if self.room.has_session():
            return HttpResponseRedirect(
                reverse('RoomWithSession', args=[kwargs['room_name']]))
        return super(RoomWithoutSession, self).dispatch(
            request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = {
            'participant_urls': self.room.get_participant_urls(self.request),
            'room_wide_url': self.room.get_room_wide_url(self.request),
            'room': self.room,
            'collapse_links': True,
        }
        kwargs.update(context)

        return super(RoomWithoutSession, self).get_context_data(**kwargs)

    def socket_url(self):
        return channel_utils.room_admin_path(self.room.name)


class RoomWithSession(vanilla.TemplateView):
    template_name = 'otree/admin/RoomWithSession.html'
    room = None

    url_pattern = r"^room_with_session/(?P<room_name>.+)/$"

    def dispatch(self, request, *args, **kwargs):
        self.room = ROOM_DICT[kwargs['room_name']]
        if not self.room.has_session():
            return HttpResponseRedirect(
                reverse('RoomWithoutSession', args=[kwargs['room_name']]))
        return super(RoomWithSession, self).dispatch(
            request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = {
            'participant_urls': self.room.get_participant_urls(self.request),
            'room_wide_url': self.room.get_room_wide_url(self.request),
            'session_url': reverse(
                'SessionMonitor',
                args=(self.room.get_session().code,)),
            'room': self.room,
            'collapse_links': True,
        }
        kwargs.update(context)

        return super(RoomWithSession, self).get_context_data(**kwargs)


class CloseRoom(vanilla.View):
    url_pattern = r"^CloseRoom/(?P<room_name>.+)/$"

    def post(self, request, *args, **kwargs):
        room_name = kwargs['room_name']
        self.room = ROOM_DICT[room_name]
        self.room.set_session(None)
        # in case any failed to be cleared through regular ws.disconnect
        ParticipantRoomVisit.objects.filter(
            room_name=room_name,
        ).delete()
        return HttpResponseRedirect(
            reverse('RoomWithoutSession', args=[room_name]))


class StaleRoomVisits(vanilla.View):

    url_pattern = r'^StaleRoomVisits/(?P<room>\w+)/$'

    def get(self, request, *args, **kwargs):
        stale_threshold = time.time() - 20
        stale_participant_labels = ParticipantRoomVisit.objects.filter(
            room_name=kwargs['room'],
            last_updated__lt=stale_threshold
        ).values_list('participant_label', flat=True)

        # make json serializable
        stale_participant_labels = list(stale_participant_labels)

        return JsonResponse({'participant_labels': stale_participant_labels})


class ActiveRoomParticipantsCount(vanilla.View):

    url_pattern = r'^ActiveRoomParticipantsCount/(?P<room>\w+)/$'

    def get(self, request, *args, **kwargs):
        count = ParticipantRoomVisit.objects.filter(
            room_name=kwargs['room'],
            last_updated__gte=time.time() - 20
        ).count()

        return JsonResponse({'count': count})


from otree.db import models

class _SaveTheChangeWithCustomFieldSupport:
    '''
    2017-08-07: kept around because old migrations files reference it.
    after a few months when i squash migrations,
    the references to this will be deleted, so i can delete it.

    2017-09-05: I found a bug with NumPy + SaveTheChange;
    https://github.com/karanlyons/django-save-the-change/issues/27
    So I need to use this again. Implementing a simplified version of what
    Gregor made a while back.
    '''

    _pickle_fields = ['vars']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._save_the_change_store_initial_pickle_fields()

    def save(self, *args, **kwargs):
        self._save_the_change_check_pickle_field_changes()
        return super().save(*args, **kwargs)

    def _save_the_change_store_initial_pickle_fields(self):
        self._initial_prep_values = {}
        for field_name in self._pickle_fields:
            field = self._meta.get_field(field_name)
            self._initial_prep_values[field_name] = field.get_prep_value(
                getattr(self, field_name))

    def _save_the_change_check_pickle_field_changes(self):
        for field_name in self._pickle_fields:
            field = self._meta.get_field(field_name)
            new_value = field.get_prep_value(getattr(self, field_name))
            initial_prep_value = self._initial_prep_values[field_name]
            if new_value != initial_prep_value:
                self._changed_fields[field_name] = field.to_python(initial_prep_value)


class ModelWithVars(_SaveTheChangeWithCustomFieldSupport, models.Model):

    class Meta:
        abstract = True

    vars = models._PickleField(default=dict)  # type: dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import channels
import json
from django.urls import reverse

from otree import constants_internal
import otree.common_internal
from otree.common_internal import (
    random_chars_8, random_chars_10, get_admin_secret_code,
    get_app_label_from_name
)
import time

from otree.db import models
from otree.models_concrete import ParticipantToPlayerLookup, RoomToSession
from django.template.loader import get_template
from django.template import TemplateDoesNotExist
from .varsmixin import ModelWithVars


logger = logging.getLogger('otree')


ADMIN_SECRET_CODE = get_admin_secret_code()

class Session(ModelWithVars):

    class Meta:
        app_label = "otree"
        # if i don't set this, it could be in an unpredictable order
        ordering = ['pk']

    _pickle_fields = ['vars', 'config']
    config = models._PickleField(default=dict, null=True)  # type: dict

    # label of this session instance
    label = models.CharField(
        max_length=300, null=True, blank=True,
        help_text='For internal record-keeping')

    experimenter_name = models.CharField(
        max_length=300, null=True, blank=True,
        help_text='For internal record-keeping')

    ready_for_browser = models.BooleanField(default=False)

    code = models.CharField(
        default=random_chars_8,
        max_length=16,
        # set non-nullable, until we make our CharField non-nullable
        null=False,
        unique=True,
        doc="Randomly generated unique identifier for the session.")

    mturk_HITId = models.CharField(
        max_length=300, null=True, blank=True,
        help_text='Hit id for this session on MTurk')
    mturk_HITGroupId = models.CharField(
        max_length=300, null=True, blank=True,
        help_text='Hit id for this session on MTurk')

    # since workers can drop out number of participants on server should be
    # greater than number of participants on mturk
    # value -1 indicates that this session it not intended to run on mturk
    mturk_num_participants = models.IntegerField(
        default=-1,
        help_text="Number of participants on MTurk")

    mturk_use_sandbox = models.BooleanField(
        default=True,
        help_text="Should this session be created in mturk sandbox?")

    # use Float instead of DateTime because DateTime
    # is a pain to work with (e.g. naive vs aware datetime objects)
    # and there is no need here for DateTime
    mturk_expiration = models.FloatField(
        null=True
    )

    archived = models.BooleanField(
        default=False,
        db_index=True,
        doc=("If set to True the session won't be visible on the "
             "main ViewList for sessions"))

    comment = models.TextField(blank=True)

    _anonymous_code = models.CharField(
        default=random_chars_10, max_length=10, null=False, db_index=True)

    _pre_create_id = models.CharField(max_length=255, db_index=True, null=True)

    def use_browser_bots(self):
        return self.participant_set.filter(is_browser_bot=True).exists()

    is_demo = models.BooleanField(default=False)

    _admin_report_app_names = models.TextField(default='')
    _admin_report_num_rounds = models.CharField(default='', max_length=255)

    num_participants = models.PositiveIntegerField()

    def __unicode__(self):
        return self.code

    @property
    def participation_fee(self):
        '''This method is deprecated from public API,
        but still useful internally (like data export)'''
        return self.config['participation_fee']

    @property
    def real_world_currency_per_point(self):
        '''This method is deprecated from public API,
        but still useful internally (like data export)'''
        return self.config['real_world_currency_per_point']

    def is_for_mturk(self):
        return (not self.is_demo) and (self.mturk_num_participants > 0)

    def get_subsessions(self):
        lst = []
        app_sequence = self.config['app_sequence']
        for app in app_sequence:
            models_module = otree.common_internal.get_models_module(app)
            subsessions = models_module.Subsession.objects.filter(
                session=self
            ).order_by('round_number')
            lst.extend(list(subsessions))
        return lst

    def get_participants(self):
        return list(self.participant_set.order_by('id_in_session'))

    def mturk_worker_url(self):
        # different HITs
        # get the same preview page, because they are lumped into the same
        # "hit group". This is not documented, but it seems HITs are lumped
        # if a certain subset of properties are the same:
        # https://forums.aws.amazon.com/message.jspa?messageID=597622#597622
        # this seems like the correct design; the only case where this will
        # not work is if the HIT was deleted from the server, but in that case,
        # the HIT itself should be canceled.


        # 2018-06-04:
        # the format seems to have changed to this:
        # https://worker.mturk.com/projects/{group_id}/tasks?ref=w_pl_prvw
        # but the old format still works.
        # it seems I can't replace groupId by hitID, which i would like to do
        # because it's more precise.
        subdomain = "workersandbox" if self.mturk_use_sandbox else 'www'
        return "https://{}.mturk.com/mturk/preview?groupId={}".format(
            subdomain,
            self.mturk_HITGroupId
        )

    def mturk_is_expired(self):
        # self.mturk_expiration is offset-aware, so therefore we must compare
        # it against an offset-aware value.
        return self.mturk_expiration and self.mturk_expiration < time.time()

    def mturk_is_active(self):
        return self.mturk_HITId and not self.mturk_is_expired()

    def advance_last_place_participants(self):
        # django.test takes 0.5 sec to import,
        # if this is done globally then it adds to each startup
        # it's only used here, and often isn't used at all.
        # so best to do it only here
        # it gets cached
        import django.test
        client = django.test.Client()

        participants = self.get_participants()

        # in case some participants haven't started
        unvisited_participants = []
        for p in participants:
            if p._index_in_pages == 0:
                unvisited_participants.append(p)
                client.get(p._start_url(), follow=True)

        if unvisited_participants:
            # that's it -- just visit the start URL, advancing by 1
            return

        last_place_page_index = min([p._index_in_pages for p in participants])
        last_place_participants = [
            p for p in participants
            if p._index_in_pages == last_place_page_index
        ]

        for p in last_place_participants:
            try:
                current_form_page_url = p._current_form_page_url
                if current_form_page_url:
                    resp = client.post(
                        current_form_page_url,
                        data={
                            constants_internal.timeout_happened: True,
                            constants_internal.admin_secret_code: ADMIN_SECRET_CODE
                        },
                        follow=True
                    )
                    # not sure why, but many users are getting HttpResponseNotFound
                    if resp.status_code >= 400:
                        msg = ('Submitting page {} failed, '
                            'returned HTTP status code {}.'.format(
                                current_form_page_url, resp.status_code
                            ))
                        content = resp.content
                        if len(content) < 600:
                            msg += ' response content: {}'.format(content)
                        raise AssertionError(msg)

                else:
                    # it's possible that the slowest user is on a wait page,
                    # especially if their browser is closed.
                    # because they were waiting for another user who then
                    # advanced past the wait page, but they were never
                    # advanced themselves.
                    start_url = p._start_url()
                    resp = client.get(start_url, follow=True)
            except:
                logging.exception("Failed to advance participants.")
                raise


            # do the auto-advancing here,
            # rather than in increment_index_in_pages,
            # because it's only needed here.
            channels.Group(
                'auto-advance-{}'.format(p.code)
            ).send(
                {'text': json.dumps(
                    {'auto_advanced': True})}
            )

    def get_room(self):
        from otree.room import ROOM_DICT
        try:
            room_name = RoomToSession.objects.get(session=self).room_name
            return ROOM_DICT[room_name]
        except RoomToSession.DoesNotExist:
            return None

    def _get_payoff_plus_participation_fee(self, payoff):
        '''For a participant who has the given payoff,
        return their payoff_plus_participation_fee
        Useful to define it here, for data export
        '''

        return (
            self.config['participation_fee'] +
            payoff.to_real_world_currency(self)
        )

    def _set_admin_report_app_names(self):

        admin_report_app_names = []
        num_rounds_list = []
        for app_name in self.config['app_sequence']:
            models_module = otree.common_internal.get_models_module(app_name)
            app_label = get_app_label_from_name(app_name)
            try:
                get_template('{}/AdminReport.html'.format(app_label))
                admin_report_app_names.append(app_name)
                num_rounds_list.append(models_module.Constants.num_rounds)
            except TemplateDoesNotExist:
                pass
        self._admin_report_app_names = ';'.join(admin_report_app_names)
        self._admin_report_num_rounds = ';'.join(str(n) for n in num_rounds_list)

    def _admin_report_apps(self):
        return self._admin_report_app_names.split(';')

    def _admin_report_num_rounds_list(self):
        return [int(num) for num in self._admin_report_num_rounds.split(';')]

    def has_admin_report(self):
        return bool(self._admin_report_app_names)



from contextlib import contextmanager
import idmap.models
import idmap.tls
import threading
import idmap

_toggle = threading.local()


def is_active():
    return getattr(_toggle, 'is_active', False)


def deactivate_cache():
    '''
    The Idmap cache is always being populated, even if it's not "active"
    We just ignore its contents.
    It doesn't look like idmap has a way to turn it off entirely.
    '''
    _toggle.is_active = False
    # 2017-08-07: flush both in activate and deactivate, just to be sure
    # i was getting some unexpected behavior in tests, when a test without
    # IDmap ran after a test with IDmap
    idmap.flush()


def activate_cache():
    idmap.flush()
    _toggle.is_active = True


@contextmanager
def use_cache():
    activate_cache()
    try:
        yield
    finally:
        deactivate_cache()


class IdMapModel(idmap.models.IdMapModel):
    class Meta:
        abstract = True
        # needs to be here for Participant and Session
        # 2018-11-24: but it seems Participant and Session don't inherit
        # from this; that's why we set use_strong_refs in the metaclass
        use_strong_refs = True

    # The ``get_cached_instance`` method is the canonical access point for
    # idmap to retrieve objects for a particular model from the cache. If it
    # returns None, then it's meant as a cache miss and the object is retrieved
    # from the database.

    # We intercept this so that we can disable the idmap cache. That is
    # required as we only want it to be active on experiment views. idmap has
    # it's problems when used together with django-channels as channels is
    # re-using threads between requests. That will result in a shared idmap
    # cache between requests, which again results in unpredictable data
    # returned from the cache as it might contain stale data.

    # The solution is to only use the cache when processing a view and clear
    # the cache before activating the use. See the activate_cache() for the
    # implementation.
    @classmethod
    def get_cached_instance(cls, *args, **kwargs):
        if is_active():
            return super().get_cached_instance(*args, **kwargs)

CLASSES_TO_SAVE = {
    'Session',
    'Participant',
    'Subsession',
    'Group',
    'Player'
}


def _get_save_objects_model_instances():
    """
    Get all model instances that should be saved. This implementation uses
    the idmap cache to determine which instances have been loaded.
    """
    # 2017-08-08: adding this because a test was failing unexpectedly for me
    # it seems reasonable that if we haven't activated the cache, we should
    # disregard it entirely.
    if not is_active():
        return []
    cache = getattr(idmap.tls._tls, 'idmap_cache', {})
    instances = []
    for model_class, model_cache in cache.items():
        # Collect instances if it's a subclass of one of the monitored
        # models.
        is_monitored = model_class.__name__ in CLASSES_TO_SAVE
        if is_monitored:
            cached_instances = list(model_cache.values())
            instances.extend(cached_instances)
    return instances


def save_objects():
    for instance in _get_save_objects_model_instances():
        # if there are no changes on the instance, SaveTheChange will detect
        # that very quickly and skip saving. I tested and it made no difference
        # if I test whether _changed_fields is empty.
        instance.save()



from importlib import import_module
from django.conf import settings
import importlib.util


"""
(THIS IS CURRENTLY PRIVATE API, MAY CHANGE WITHOUT NOTICE)

To create an oTree extension, add a package called ``otree_extensions``
to your app, and add the app name in settings.py to EXTENSION_APPS. 
It can contain any of the following submodules:

urls.py
-------

should contain a variable ``urlpatterns``, which will be appended to
oTree's built-in URL patterns.

routing.py
----------

Should contain a variable ``channel_routing``,
with a list of channel routes, as described in the Django channels documentation:

https://channels.readthedocs.io/en/stable/getting-started.html#routing

These routes will be appended to oTree's built-in channel routes.

admin.py
--------

This module allows you to define custom data exports that will be included
in oTree's data export page. Define a variable ``data_export_views``,
which is a list of Django class-based views (see Django docs).

Each view should define a ``get()`` method with the following signature::

    def get(self, request, *args, **kwargs):

This method should return an HTTP response with
the exported data (e.g. CSV, XLSX, JSON, etc), using the appropriate MIME type
on the HTTP response.

Each view must also have the following attributes:

-   ``url_pattern``: the URL pattern string, e.g. '^mychat_export/$'
-   ``url_name``: see Django docs on reverse resolution of URLs, e.g. 'mychat_export'
-   ``display_name``: The text of the download hyperlink on the data export page
    (e.g. "Chat Data Export")

You don't need to worry about login_required and AUTH_LEVEL;
oTree will handle this automatically.

(In the future, admin.py may be used for other admin customizations,
not just data export.)

"""

def get_extensions_modules(submodule_name):
    modules = []
    extension_apps = getattr(settings, 'EXTENSION_APPS', [])
    # legacy support for otreechat
    if 'otreechat' in settings.INSTALLED_APPS:
        extension_apps.append('otreechat')
    find_spec = importlib.util.find_spec
    for app_name in extension_apps:
        package_dotted = '{}.otree_extensions'.format(app_name)
        submodule_dotted = '{}.{}'.format(package_dotted, submodule_name)
        # need to check if base package exists; otherwise we get ImportError
        if find_spec(package_dotted) and find_spec(submodule_dotted):
            module = import_module(submodule_dotted)
            modules.append(module)
    return modules


def get_extensions_data_export_views():
    view_classes = []
    for module in get_extensions_modules('admin'):
        view_classes += getattr(module, 'data_export_views', [])
    return view_classes


import json
import os
import sys
from collections import OrderedDict


import channels
import otree.bots.browser
import otree.channels.utils as channel_utils
import otree.common_internal
import otree.export
import otree.models
import vanilla
from django.conf import settings
from django.contrib import messages
from django.urls import reverse
from django.http import HttpResponseRedirect, JsonResponse, HttpResponse, \
    Http404
from django.shortcuts import get_object_or_404
from otree import forms
from otree.common import RealWorldCurrency
from otree.common_internal import (
    create_session_and_redirect, missing_db_tables,
    get_models_module, get_app_label_from_name, DebugTable,
)
from otree.forms import widgets
from otree_startup import check_pypi_for_updates
from otree.models import Participant, Session
from otree.models_concrete import (
    BrowserBotsLauncherSessionCode)
from otree.session import SESSION_CONFIGS_DICT, create_session, SessionConfig
from otree.views.abstract import GenericWaitPageMixin, AdminSessionPageMixin
from django.db.models import Case, Value, When

def pretty_name(name):
    """Converts 'first_name' to 'first name'"""
    if not name:
        return ''
    return name.replace('_', ' ')


class CreateSessionForm(forms.Form):
    session_configs = SESSION_CONFIGS_DICT.values()
    session_config_choices = (
        # use '' instead of None. '' seems to immediately invalidate the choice,
        # rather than None which seems to be coerced to 'None'.
        [('', '-----')] +
        [(s['name'], s['display_name']) for s in session_configs])

    session_config = forms.ChoiceField(
        choices=session_config_choices, required=True)

    num_participants = forms.IntegerField()

    def __init__(self, *args, **kwargs):
        for_mturk = kwargs.pop('for_mturk')
        super().__init__(*args, **kwargs)
        if for_mturk:
            self.fields['num_participants'].label = "Number of MTurk workers"
            self.fields['num_participants'].help_text = (
                'Since workers can return the HIT or drop out, '
                'some "spare" participants will be created: '
                'the oTree session will have '
                '{} times more participants than the MTurk HIT. '
                'The number you enter in this field is number of '
                'workers required for your HIT.'.format(
                    settings.MTURK_NUM_PARTICIPANTS_MULTIPLE
                )
            )
        else:
            self.fields['num_participants'].label = "Number of participants"

    def clean_num_participants(self):
        session_config_name = self.cleaned_data.get('session_config')

        # I think when this is checked, it's possible that basic validation
        # for session_config_name was not done yet.
        # when I tested it was None
        # but maybe it could also be the empty string because that's what's
        # explicitly put above.
        if session_config_name:
            lcm = SESSION_CONFIGS_DICT[session_config_name].get_lcm()
            num_participants = self.cleaned_data['num_participants']
            if num_participants % lcm:
                raise forms.ValidationError(
                    'Please enter a valid number of participants.'
                )
            return num_participants


class CreateSession(vanilla.FormView):
    form_class = CreateSessionForm
    template_name = 'otree/admin/CreateSession.html'

    url_pattern = r"^create_session/$"

    def dispatch(self, request, *args, **kwargs):
        # splinter makes request.GET.get('mturk') == ['1\\']
        # no idea why
        # so just see if it's non-empty
        self.for_mturk = bool(self.request.GET.get('mturk'))
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs['configs'] = SESSION_CONFIGS_DICT.values()
        return super().get_context_data(**kwargs)

    def get_form(self, data=None, files=None, **kwargs):
        kwargs['for_mturk'] = self.for_mturk
        return super().get_form(data, files, **kwargs)

    def form_valid(self, form):

        session_config_name = form.cleaned_data['session_config']
        session_kwargs = {
            'session_config_name': session_config_name,
            'for_mturk': self.for_mturk
        }
        if self.for_mturk:
            session_kwargs['num_participants'] = (
                form.cleaned_data['num_participants'] *
                settings.MTURK_NUM_PARTICIPANTS_MULTIPLE
            )

        else:
            session_kwargs['num_participants'] = (
                form.cleaned_data['num_participants'])

        # TODO:
        # Refactor when we upgrade to push
        if hasattr(self, "room"):
            session_kwargs['room_name'] = self.room.name

        post_data = self.request.POST
        config = SESSION_CONFIGS_DICT[session_config_name]

        edited_session_config_fields = {}

        for field in config.editable_fields():
            old_value = config[field]
            html_field_name = config.html_field_name(field)

            # int/float/decimal are set to required in HTML
            # bool:
            # - if unchecked, its key will be missing.
            # - if checked, its value will be 'on'.
            # str:
            # - if blank, its value will be ''
            new_value_str = post_data.get(html_field_name)
            # don't use isinstance because that will catch bool also
            if type(old_value) is int:
                # in case someone enters 1.0 instead of 1
                new_value = int(float(new_value_str))
            else:
                new_value = type(old_value)(new_value_str)
            if old_value != new_value:
                edited_session_config_fields[field] = new_value

        # need to convert to float or string in order to serialize
        # through channels
        for k in ['participation_fee', 'real_world_currency_per_point']:
            if k in edited_session_config_fields:
                edited_session_config_fields[k] = float(
                    edited_session_config_fields[k])
        session_kwargs['edited_session_config_fields'] = edited_session_config_fields


        use_browser_bots = edited_session_config_fields.get('use_browser_bots')
        if use_browser_bots is None:
            use_browser_bots = config.get('use_browser_bots', False)

        return create_session_and_redirect(
            session_kwargs, use_browser_bots=use_browser_bots)


class WaitUntilSessionCreated(GenericWaitPageMixin, vanilla.GenericView):

    url_pattern = r"^WaitUntilSessionCreated/(?P<pre_create_id>.+)/$"

    title_text = 'creating_session'
    body_text = ''

    def _is_ready(self):
        try:
            self.session = Session.objects.get(
                _pre_create_id=self._pre_create_id,
                ready_for_browser=True
            )
            return True
        except Session.DoesNotExist:
            return False

    def _response_when_ready(self):
        session = self.session
        if session.is_for_mturk():
            session_home_url = reverse(
                'MTurkCreateHIT', args=(session.code,)
            )
        # 2017-09-30: deleted a line about split screen here, because
        # the only way to get split screen is to first open the session links
        # page, then switch to split-screen mode
        else:
            session_home_url = reverse(
                'SessionStartLinks', args=(session.code,))

        return HttpResponseRedirect(session_home_url)

    def dispatch(self, request, *args, **kwargs):
        self._pre_create_id = kwargs['pre_create_id']
        if self._is_ready():
            return self._response_when_ready()
        return self._get_wait_page()

    def socket_url(self):
        return channel_utils.wait_for_session_path(self._pre_create_id)


class SessionSplitScreen(AdminSessionPageMixin, vanilla.TemplateView):
    '''Launch the session in fullscreen mode
    only used in demo mode
    '''

    def get_context_data(self, **kwargs):
        '''Get the URLs for the IFrames'''
        context = super(SessionSplitScreen, self).get_context_data(**kwargs)
        participant_urls = [
            self.request.build_absolute_uri(participant._start_url())
            for participant in self.session.get_participants()
        ]
        context.update({
            'session': self.session,
            'participant_urls': participant_urls
        })
        return context


class SessionStartLinks(AdminSessionPageMixin, vanilla.TemplateView):

    def get_context_data(self, **kwargs):
        session = self.session
        room = session.get_room()

        context = super().get_context_data(**kwargs)

        sqlite = settings.DATABASES['default']['ENGINE'].endswith('sqlite3')
        context.update({
            'use_browser_bots': session.use_browser_bots(),
            'sqlite': sqlite,
            'runserver': 'runserver' in sys.argv or 'devserver' in sys.argv
        })

        session_start_urls = [
            self.request.build_absolute_uri(participant._start_url())
            for participant in session.get_participants()
        ]

        # TODO: Bot URLs, and a button to start the bots

        if room:
            context.update(
                {
                    'participant_urls':
                        room.get_participant_urls(self.request),
                    'room_wide_url': room.get_room_wide_url(self.request),
                    'session_start_urls': session_start_urls,
                    'room': room,
                    'collapse_links': True,
                })
        else:
            anonymous_url = self.request.build_absolute_uri(
                reverse(
                    'JoinSessionAnonymously',
                    args=(session._anonymous_code,)
                )
            )

            context.update({
                'participant_urls': session_start_urls,
                'anonymous_url': anonymous_url,
                'num_participants': len(session_start_urls),
                'splitscreen_mode_on': len(session_start_urls) <= 3
            })

        return context


class SessionEditPropertiesForm(forms.ModelForm):
    participation_fee = forms.RealWorldCurrencyField(
        required=False,
        # it seems that if this is omitted, the step defaults to an integer,
        # meaninng fractional inputs are not accepted
        widget=widgets._RealWorldCurrencyInput(attrs={'step': 0.01})
    )
    real_world_currency_per_point = forms.FloatField(
        required=False
    )

    class Meta:
        model = Session
        fields = [
            'label',
            'experimenter_name',
            'comment',
        ]


class SessionEditProperties(AdminSessionPageMixin, vanilla.UpdateView):

    # required for vanilla.UpdateView
    lookup_field = 'code'
    model = Session
    form_class = SessionEditPropertiesForm
    template_name = 'otree/admin/SessionEditProperties.html'

    def get_form(self, data=None, files=None, **kwargs):
        form = super(
            SessionEditProperties, self
        ).get_form(data, files, **kwargs)
        config = self.session.config
        form.fields[
            'participation_fee'
        ].initial = config['participation_fee']
        form.fields[
            'real_world_currency_per_point'
        ].initial = config['real_world_currency_per_point']
        if self.session.mturk_HITId:
            form.fields['participation_fee'].widget.attrs['readonly'] = 'True'
        return form

    def get_success_url(self):
        return reverse('SessionEditProperties', args=(self.session.code,))

    def form_valid(self, form):
        super(SessionEditProperties, self).form_valid(form)
        participation_fee = form.cleaned_data[
            'participation_fee'
        ]
        real_world_currency_per_point = form.cleaned_data[
            'real_world_currency_per_point'
        ]
        config = self.session.config
        if form.cleaned_data['participation_fee'] is not None:
            config[
                'participation_fee'
            # need to convert back to RealWorldCurrency, because easymoney
            # MoneyFormField returns a decimal, not Money (not sure why)
            ] = RealWorldCurrency(participation_fee)
        if form.cleaned_data['real_world_currency_per_point'] is not None:
            config[
                'real_world_currency_per_point'
            ] = real_world_currency_per_point
        self.session.save()
        messages.success(self.request, 'Properties have been updated')
        return HttpResponseRedirect(self.get_success_url())


class SessionPayments(AdminSessionPageMixin, vanilla.TemplateView):

    def get(self, *args, **kwargs):
        response = super(SessionPayments, self).get(*args, **kwargs)
        return response

    def get_context_data(self, **kwargs):
        session = self.session
        # TODO: mark which ones are bots
        participants = session.get_participants()
        total_payments = 0.0
        mean_payment = 0.0
        if participants:
            total_payments = sum(
                part.payoff_plus_participation_fee() for part in participants
            )
            mean_payment = total_payments / len(participants)

        context = super(SessionPayments, self).get_context_data(**kwargs)
        context.update({
            'participants': participants,
            'total_payments': total_payments,
            'mean_payment': mean_payment,
            'participation_fee': session.config['participation_fee'],
        })

        return context


def pretty_round_name(app_label, round_number):
    app_label = pretty_name(app_label)
    if round_number > 1:
        return '{} [Round {}]'.format(app_label, round_number)
    else:
        return app_label


class SessionData(AdminSessionPageMixin, vanilla.TemplateView):

    def get_context_data(self, **kwargs):
        session = self.session

        rows = []

        round_headers = []
        model_headers = []
        field_names = []

        # field names for JSON response
        field_names_json = []

        for subsession in session.get_subsessions():
            # can't use subsession._meta.app_config.name, because it won't work
            # if the app is removed from SESSION_CONFIGS after the session is
            # created.
            columns_for_models, subsession_rows = otree.export.get_rows_for_live_update(subsession)

            if not rows:
                rows = subsession_rows
            else:
                for i in range(len(rows)):
                    rows[i].extend(subsession_rows[i])

            round_colspan = 0
            for model_name in ['player', 'group', 'subsession']:
                colspan = len(columns_for_models[model_name])
                model_headers.append((model_name.title(), colspan))
                round_colspan += colspan

            round_name = pretty_round_name(subsession._meta.app_label, subsession.round_number)

            round_headers.append((round_name, round_colspan))

            this_round_fields = []
            this_round_fields_json = []
            for model_name in ['Player', 'Group', 'Subsession']:
                column_names = columns_for_models[model_name.lower()]
                this_model_fields = [pretty_name(n) for n in column_names]
                this_model_fields_json = [
                    '{}.{}.{}'.format(round_name, model_name, colname)
                    for colname in column_names
                ]
                this_round_fields.extend(this_model_fields)
                this_round_fields_json.extend(this_model_fields_json)

            field_names.extend(this_round_fields)
            field_names_json.extend(this_round_fields_json)

        # dictionary for json response
        # will be used only if json request  is done

        self.context_json = []
        for i, row in enumerate(rows, start=1):
            d_row = OrderedDict()
            # table always starts with participant 1
            d_row['participant_label'] = 'P{}'.format(i)
            for t, v in zip(field_names_json, row):
                d_row[t] = v
            self.context_json.append(d_row)

        context = super().get_context_data(**kwargs)
        context.update({
            'subsession_headers': round_headers,
            'model_headers': model_headers,
            'field_headers': field_names,
            'rows': rows})
        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        if self.request.META.get('CONTENT_TYPE') == 'application/json':
            return JsonResponse(self.context_json, safe=False)
        else:
            return self.render_to_response(context)


class SessionMonitor(AdminSessionPageMixin, vanilla.TemplateView):

    def get_context_data(self, **kwargs):

        field_names = otree.export.get_field_names_for_live_update(Participant)
        display_names = {
            '_id_in_session': 'ID in session',
            'code': 'Code',
            'label': 'Label',
            '_current_page': 'Page',
            '_current_app_name': 'App',
            '_round_number': 'Round',
            '_current_page_name': 'Page name',
            'status': 'Status',
            '_last_page_timestamp': 'Time on page',
        }

        callable_fields = {'status', '_id_in_session', '_current_page'}

        column_names = [display_names[col] for col in field_names]

        context = super().get_context_data(**kwargs)
        context['column_names'] = column_names

        advance_users_button_text = (
            "Advance the slowest user(s) by one page, "
            "by forcing a timeout on their current page. "
        )
        context['advance_users_button_text'] = advance_users_button_text


        participants = self.session.participant_set.filter(visited=True)
        rows = []

        for participant in participants:
            row = {}
            for field_name in field_names:
                value = getattr(participant, field_name)
                if field_name in callable_fields:
                    value = value()
                row[field_name] = value
            rows.append(row)

        self.context_json = rows

        return context

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        if self.request.META.get('CONTENT_TYPE') == 'application/json':
            return JsonResponse(self.context_json, safe=False)
        else:
            return self.render_to_response(context)



class SessionDescription(AdminSessionPageMixin, vanilla.TemplateView):

    def get_context_data(self, **kwargs):
        context = super(SessionDescription, self).get_context_data(**kwargs)
        context['config'] = SessionConfig(self.session.config)
        return context


class AdminReportForm(forms.Form):
    app_name = forms.ChoiceField(choices=[], required=False)
    round_number = forms.IntegerField(required=False, min_value=1)

    def __init__(self, *args, **kwargs):
        self.session = kwargs.pop('session')
        super().__init__(*args, **kwargs)

        admin_report_apps = self.session._admin_report_apps()
        num_rounds_list = self.session._admin_report_num_rounds_list()
        self.rounds_per_app = dict(zip(admin_report_apps, num_rounds_list))
        app_name_choices = []
        for app_name in admin_report_apps:
            label = '{} ({} rounds)'.format(
                get_app_label_from_name(app_name), self.rounds_per_app[app_name]
            )
            app_name_choices.append((app_name, label))

        self.fields['app_name'].choices = app_name_choices

    def clean(self):
        cleaned_data = super().clean()

        apps_with_admin_report = self.session._admin_report_apps()

        # can't use setdefault because the key will always exist even if the
        # fields were empty.
        # str default value is '',
        # and int default value is None
        if not cleaned_data['app_name']:
            cleaned_data['app_name'] = apps_with_admin_report[0]

        rounds_in_this_app = self.rounds_per_app[cleaned_data['app_name']]

        round_number = cleaned_data['round_number']

        if not round_number or round_number > rounds_in_this_app:
            cleaned_data['round_number'] = rounds_in_this_app

        self.data = cleaned_data

        return cleaned_data


class AdminReport(AdminSessionPageMixin, vanilla.TemplateView):

    def get(self, request, *args, **kwargs):
        form = AdminReportForm(data=request.GET, session=self.session)
        # validate to get error messages
        form.is_valid()
        context = self.get_context_data(form=form)
        return self.render_to_response(context)

    def get_context_data(self, **kwargs):

        cleaned_data = kwargs['form'].cleaned_data

        models_module = get_models_module(cleaned_data['app_name'])
        subsession = models_module.Subsession.objects.get(
            session=self.session,
            round_number=cleaned_data['round_number'],
        )

        context = {
            'subsession': subsession,
            'Constants': models_module.Constants,
            'session': self.session,
            'user_template': '{}/AdminReport.html'.format(
            subsession._meta.app_config.label)
        }

        vars_for_admin_report = subsession.vars_for_admin_report() or {}
        self.debug_tables = [
            DebugTable(
                title='vars_for_admin_report',
                rows=vars_for_admin_report.items()
            )
        ]
        # determine whether to display debug tables
        self.is_debug = settings.DEBUG
        context.update(vars_for_admin_report)

        # this should take priority, in the event of a clash between
        # a user-defined var and a built-in one
        context.update(super().get_context_data(**kwargs))



        return context


class ServerCheck(vanilla.TemplateView):
    template_name = 'otree/admin/ServerCheck.html'

    url_pattern = r"^server_check/$"

    def app_is_on_heroku(self):
        return 'heroku' in self.request.get_host()

    def worker_is_running(self):
        if otree.common_internal.USE_REDIS:
            redis_conn = otree.common_internal.get_redis_conn()
            return otree.bots.browser.ping_bool(redis_conn, timeout=2)
        else:
            # the timeoutworker relies on Redis (Huey),
            # so if Redis is not being used, the timeoutworker is not functional
            return False

    def get_context_data(self, **kwargs):
        sqlite = settings.DATABASES['default']['ENGINE'].endswith('sqlite3')
        debug = settings.DEBUG
        regular_sentry = hasattr(settings, 'RAVEN_CONFIG')
        # TODO:
        # is this line necessary?
        # how does Heroku's Sentry addon report errors without Raven?
        heroku_sentry = os.environ.get('SENTRY_DSN')
        sentry = regular_sentry or heroku_sentry
        auth_level = settings.AUTH_LEVEL
        auth_level_ok = settings.AUTH_LEVEL in {'DEMO', 'STUDY'}
        heroku = self.app_is_on_heroku()
        runserver = ('runserver' in sys.argv) or ('devserver' in sys.argv)
        db_synced = not missing_db_tables()
        pypi_results = check_pypi_for_updates()
        worker_is_running = self.worker_is_running()

        return {
            'sqlite': sqlite,
            'debug': debug,
            'sentry': sentry,
            'auth_level': auth_level,
            'auth_level_ok': auth_level_ok,
            'heroku': heroku,
            'runserver': runserver,
            'db_synced': db_synced,
            'pypi_results': pypi_results,
            'worker_is_running': worker_is_running,
        }


class OtreeCoreUpdateCheck(vanilla.View):

    url_pattern = r"^version_cached/$"

    # cached per process
    results = None

    def get(self, request, *args, **kwargs):
        if OtreeCoreUpdateCheck.results is None:
            OtreeCoreUpdateCheck.results = check_pypi_for_updates()
        return JsonResponse(OtreeCoreUpdateCheck.results, safe=True)


class CreateBrowserBotsSession(vanilla.View):

    url_pattern = r"^create_browser_bots_session/$"

    def get(self, request, *args, **kwargs):
        # return browser bots check
        sqlite = settings.DATABASES['default']['ENGINE'].endswith('sqlite3')

        return JsonResponse({
            'sqlite': sqlite,
            'runserver': 'runserver' in sys.argv or 'devserver' in sys.argv
        })

    def post(self, request, *args, **kwargs):
        num_participants = int(request.POST['num_participants'])
        session_config_name = request.POST['session_config_name']
        case_number = int(request.POST['case_number'])
        session = create_session(
            session_config_name=session_config_name,
            num_participants=num_participants,
        )
        otree.bots.browser.initialize_session(
            session_pk=session.pk, case_number=case_number)
        BrowserBotsLauncherSessionCode.objects.update_or_create(
            # i don't know why the update_or_create arg is called 'defaults'
            # because it will update even if the instance already exists
            # maybe for consistency with get_or_create
            defaults={'code': session.code}
        )
        channels.Group('browser_bot_wait').send(
            {'text': json.dumps({'status': 'session_ready'})}
        )

        return HttpResponse(session.code)


class CloseBrowserBotsSession(vanilla.View):

    url_pattern = r"^close_browser_bots_session/$"

    def post(self, request, *args, **kwargs):
        BrowserBotsLauncherSessionCode.objects.all().delete()
        return HttpResponse('ok')


class AdvanceSession(vanilla.View):

    url_pattern = r'^AdvanceSession/(?P<session_code>[a-z0-9]+)/$'

    def post(self, *args, **kwargs):
        session = get_object_or_404(
            otree.models.Session, code=kwargs['session_code']
        )
        session.advance_last_place_participants()
        return HttpResponse('ok')


class Sessions(vanilla.ListView):
    template_name = 'otree/admin/Sessions.html'

    url_pattern = r"^sessions/$"

    def dispatch(self, request, *args, **kwargs):
        self.is_archive = self.request.GET.get('archived') == '1'
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        archived_sessions_exist = Session.objects.filter(archived=True).exists()
        context.update({
            'is_archive': self.is_archive,
            'is_debug': settings.DEBUG,
            'archived_sessions_exist': archived_sessions_exist
        })
        return context

    def get_queryset(self):
        return Session.objects.filter(
            is_demo=False, archived=self.is_archive).order_by('-pk')


class ToggleArchivedSessions(vanilla.View):

    url_pattern = r'^ToggleArchivedSessions/'

    def post(self, request, *args, **kwargs):
        code_list = request.POST.getlist('session')

        (Session.objects.filter(code__in=code_list)
            .update(archived=Case(
                When(archived=True, then=Value(False)),
                default=Value(True))
            )
        )

        return HttpResponseRedirect(reverse('Sessions'))


class DeleteSessions(vanilla.View):

    url_pattern = r'^DeleteSessions/'

    def dispatch(self, *args, **kwargs):
        return super(DeleteSessions, self).dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        for code in request.POST.getlist('session'):
            session = get_object_or_404(
                otree.models.Session, code=code
            )
            session.delete()
        return HttpResponseRedirect(reverse('Sessions'))


import logging
import six

from django.conf import settings
from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.db import connections, transaction
from otree import common_internal


logger = logging.getLogger('otree')


def drop_tables_command(db_engine):
    if 'sqlite3' in db_engine:
        return 'DROP TABLE {table};'
    if 'oracle' in db_engine:
        return 'DROP TABLE "{table}" CASCADE CONSTRAINTS;'
    if 'postgres' in db_engine:
        return 'DROP TABLE "{table}" CASCADE;'
    if 'mysql' in db_engine:
        return (
            'SET FOREIGN_KEY_CHECKS = 0;'
            'DROP TABLE {table} CASCADE;'
            'SET FOREIGN_KEY_CHECKS = 1;')
    raise ValueError(
        'resetdb command does not recognize DB engine "{}"'.format(db_engine))


class Command(BaseCommand):
    help = (
        "Resets your development database to a fresh state. "
        "All data will be deleted.")

    def add_arguments(self, parser):
        ahelp = (
            'Tells the resetdb command to NOT prompt the user for '
            'input of any kind.')
        parser.add_argument(
            '--noinput', action='store_false', dest='interactive',
            default=True, help=ahelp)

    def _confirm(self):
        self.stdout.write(
            "This will delete and recreate your database. ")
        answer = six.moves.input("Proceed? (y or n): ")
        if answer:
            return answer[0].lower() == 'y'
        return False

    def _drop_table_stmt(self, dbconf):
        engine = dbconf["ENGINE"]
        return drop_tables_command(engine)

    def _get_tables(self, db):
        tables = []
        out = six.StringIO()
        call_command('inspectdb', database=db, no_color=True, stdout=out)
        for line in out.getvalue().splitlines():
            line = line.strip()
            if line.startswith("db_table = '"):
                tablename = line.replace(
                    "db_table = '", "", 1).replace("'", "").strip()
                tables.append(tablename)
        return tuple(reversed(tables))

    def _drop_tables(self, tables, db, dt_stmt):
        with connections[db].cursor() as cursor:
            for table in tables:
                stmt = dt_stmt.format(table=table)
                cursor.execute(stmt)

    def handle(self, **options):
        if options.pop("interactive") and not self._confirm():
            self.stdout.write('Canceled.')
            return

        for db, dbconf in six.iteritems(settings.DATABASES):
            db_engine = dbconf['ENGINE']
            if 'postgresql' in db_engine.lower():
                db_engine = 'PostgreSQL'
            elif 'sqlite' in db_engine.lower():
                db_engine = 'SQLite'
            elif 'mysql' in db_engine.lower():
                db_engine = 'MySQL'
            # hub depends on this string
            logger.info("Database engine: {}".format(db_engine))
            dt_stmt = self._drop_table_stmt(dbconf)

            logger.info("Retrieving Existing Tables...")
            tables = self._get_tables(db)

            logger.info("Dropping Tables...")

            # use a transaction to prevent the DB from getting in an erroneous
            # state, which can result in a different error message when resetdb
            # is run again, making the original error hard to trace.
            with transaction.atomic(
                    using=connections[db].alias,
                    savepoint=connections[db].features.can_rollback_ddl
            ):
                self._drop_tables(tables, db, dt_stmt)

                logger.info("Creating Database '{}'...".format(db))

                # need to hide the migrations/ folder so that Django thinks
                # it doesn't exist.
                # Tried setting MIGRATIONS_MODULES but doesn't work
                # (causes ModuleNotFoundError)
                common_internal.patch_migrations_module()

                call_command(
                    'migrate', database=db,
                    interactive=False, run_syncdb=True, **options)

        # mention the word 'columns' here, so people make the connection
        # between columns and resetdb, so that when they get a 'no such column'
        # error, they know how to fix it.
        # (An alternative is to generically catch "no such column" errors,
        # but I recall that this was difficult - because there were many
        # code paths or exception classes. Could re-investigate.)

        # 2018-11-08: oTree Hub depends on this string
        logger.info('Created new tables and columns.')



import logging
from django.core.management import call_command
from . import webandworkers

logger = logging.getLogger(__name__)


class Command(webandworkers.Command):

    '''
    It's almost the same as webandworkers,
    but it also runs collectstatic,
    and launches the timeoutworker.
    this command is intended for generic server deployment,
    whereas webandworkers is intended for Heroku.
    '''

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        ahelp = (
            'By default we will collect all static files into the directory '
            'configured in your settings. Disable it with this switch if you '
            'want to do it manually.')
        parser.add_argument(
            '--no-collectstatic', action='store_false', dest='collectstatic',
            default=True, help=ahelp)

    def setup_honcho(self, options):
        super().setup_honcho(options)
        honcho = self.honcho

        honcho.add_otree_process(
            'botworker',
            'otree botworker',
        )
        honcho.add_otree_process(
            'timeoutworkeronly',
            'otree timeoutworkeronly',
        )

    def handle(self, *args, **options):
        collectstatic = options['collectstatic']

        if collectstatic:
            self.stdout.write('Running collectstatic ...')
            call_command('collectstatic', interactive=False, verbosity=1)

        return super(Command, self).handle(*args, **options)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six
from django.db.models import Prefetch
from otree.db import models
from otree.common_internal import (
    get_models_module, in_round, in_rounds)
from otree import matching
import copy
from collections import defaultdict
from otree.common_internal import has_group_by_arrival_time


class GroupMatrixError(ValueError):
    pass


class RoundMismatchError(GroupMatrixError):
    pass

class BaseSubsession(models.Model):
    """Base class for all Subsessions.
    """

    class Meta:
        abstract = True
        ordering = ['pk']
        index_together = ['session', 'round_number']

    session = models.ForeignKey(
        'otree.Session', related_name='%(app_label)s_%(class)s', null=True,
        on_delete=models.CASCADE,
    )

    round_number = models.PositiveIntegerField(
        db_index=True,
        doc='''If this subsession is repeated (i.e. has multiple rounds), this
        field stores the position of this subsession, among subsessions
        in the same app.
        '''
    )

    def in_round(self, round_number):
        return in_round(type(self), round_number,
            session=self.session,
        )

    def in_rounds(self, first, last):
        return in_rounds(type(self), first, last, session=self.session)

    def in_previous_rounds(self):
        return self.in_rounds(1, self.round_number-1)

    def in_all_rounds(self):
        return self.in_previous_rounds() + [self]

    def __unicode__(self):
        return str(self.pk)

    def get_groups(self):
        return list(self.group_set.order_by('id_in_subsession'))

    def get_players(self):
        return list(self.player_set.order_by('pk'))

    def get_group_matrix(self):
        players_prefetch = Prefetch(
            'player_set',
            queryset=self._PlayerClass().objects.order_by('id_in_group'),
            to_attr='_ordered_players')
        return [group._ordered_players
                for group in self.group_set.order_by('id_in_subsession')
                                 .prefetch_related(players_prefetch)]

    def set_group_matrix(self, matrix):
        """
        warning: this deletes the groups and any data stored on them
        TODO: we should indicate this in docs
        """

        try:
            players_flat = [p for g in matrix for p in g]
        except TypeError:
            raise GroupMatrixError(
                'Group matrix must be a list of lists.'
            ) from None
        try:
            matrix_pks = sorted(p.pk for p in players_flat)
        except AttributeError:
            # if integers, it's OK
            if isinstance(players_flat[0], int):
                # deep copy so that we don't modify the input arg
                matrix = copy.deepcopy(matrix)
                players_flat = sorted(players_flat)
                players_from_db = self.get_players()
                if players_flat == list(range(1, len(players_from_db) + 1)):
                    for i, row in enumerate(matrix):
                        for j, val in enumerate(row):
                            matrix[i][j] = players_from_db[val - 1]
                else:
                    raise GroupMatrixError(
                        'If you pass a matrix of integers to this function, '
                        'It must contain all integers from 1 to '
                        'the number of players in the subsession.'
                    ) from None
            else:
                raise GroupMatrixError(
                    'The elements of the group matrix '
                    'must either be Player objects, or integers.'
                ) from None
        else:
            existing_pks = list(
                self.player_set.values_list(
                    'pk', flat=True
                ).order_by('pk'))
            if matrix_pks != existing_pks:
                wrong_round_numbers = [
                    p.round_number for p in players_flat
                    if p.round_number != self.round_number]
                if wrong_round_numbers:
                    raise GroupMatrixError(
                        'You are setting the groups for round {}, '
                        'but the matrix contains players from round {}.'.format(
                            self.round_number,
                            wrong_round_numbers[0]
                        )
                    )
                raise GroupMatrixError(
                    'The group matrix must contain each player '
                    'in the subsession exactly once.'
                )

        # Before deleting groups, Need to set the foreignkeys to None
        # 2016-11-8: does this need to be in a transaction?
        # because what if a player refreshes their page while this is going
        # on?
        self.player_set.update(group=None)
        self.group_set.all().delete()

        GroupClass = self._GroupClass()
        for i, row in enumerate(matrix, start=1):
            group = GroupClass.objects.create(
                subsession=self, id_in_subsession=i,
                session=self.session, round_number=self.round_number)

            group.set_players(row)

    def group_like_round(self, round_number):
        previous_round = self.in_round(round_number)
        group_matrix = [
            group._ordered_players
            for group in previous_round.group_set.order_by(
                'id_in_subsession'
            ).prefetch_related(
                Prefetch(
                    'player_set',
                    queryset=self._PlayerClass().objects.order_by(
                        'id_in_group'),
                    to_attr='_ordered_players'
                )
            )
        ]
        for i, group_list in enumerate(group_matrix):
            for j, player in enumerate(group_list):
                # for every entry (i,j) in the matrix, follow the pointer
                # to the same person in the next round
                group_matrix[i][j] = player.in_round(self.round_number)

        self.set_group_matrix(group_matrix)

    def new_group_like_round(self, round_number):
        '''test this, could work'''
        matrix = self.in_round(round_number).get_group_matrix()
        for row in matrix:
            for col in row:
                matrix[row][col] = matrix[row][col].id_in_subsession
        self.set_group_matrix(matrix)

    '''
    def group_like_round(self, round_number):
        PlayerClass = self._PlayerClass()
        last_round_info = PlayerClass.objects.filter(
            session_id=self.session_id,
            round_number=round_number
        ).values(
            'id_in_group', 'participant_id', 'group__id_in_subsession'
        ).order_by('group__id_in_subsession', 'id_in_group')

        player_lookups = {p.participant_id: p for p in self.get_players()}

        self.player_set.update(group=None)
        self.group_set.all().delete()

        # UNFINISHED
        GroupClass = self._GroupClass()
        for i, row in enumerate(matrix, start=1):
            group = GroupClass.objects.create(
                subsession=self, id_in_subsession=i,
                session=self.session, round_number=self.round_number)

            group.set_players(row)
    '''


    def set_groups(self, matrix):
        '''renamed this to set_group_matrix, but keeping in for compat'''
        return self.set_group_matrix(matrix)

    @property
    def _Constants(self):
        return get_models_module(self._meta.app_config.name).Constants

    def _GroupClass(self):
        return models.get_model(self._meta.app_config.label, 'Group')

    def _PlayerClass(self):
        return models.get_model(self._meta.app_config.label, 'Player')

    @classmethod
    def _has_group_by_arrival_time(cls):
        return has_group_by_arrival_time(cls._meta.app_config.name)


    def group_randomly(self, *, fixed_id_in_group=False):
        group_matrix = self.get_group_matrix()
        group_matrix = matching.randomly(
            group_matrix,
            fixed_id_in_group)
        self.set_group_matrix(group_matrix)

    def _group_by_rank(self, ranked_list):
        # FIXME: delete this
        group_matrix = matching.by_rank(
            ranked_list,
            self._Constants.players_per_group
        )
        self.set_group_matrix(group_matrix)

    def before_session_starts(self):
        '''Deprecated and renamed to creating_session'''
        pass

    def creating_session(self):
        pass

    def vars_for_admin_report(self):
        return {}


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# NOTE: this imports the following submodules and then subclasses several
# classes importing is done via import_module rather than an ordinary import.
#
# The only reason for this is to hide the base classes from IDEs like PyCharm,
# so that those members/attributes don't show up in autocomplete,
# including all the built-in django fields that an ordinary oTree programmer
# will never need or want. if this was a conventional Django project I wouldn't
# do it this way, but because oTree is aimed at newcomers who may need more
# assistance from their IDE, I want to try this approach out.
#
# This module is also a form of documentation of the public API.

# 2016-07-18: not using the import_module trick for now, because currently,
# the PlayerBot class doesn't have any methods we need to hide
# from importlib import import_module
# otree_bot = import_module('otree.bots.bot')

from importlib import import_module

_bot_module = import_module('otree.bots.bot')

Bot = _bot_module.PlayerBot
Submission = _bot_module.Submission
SubmissionMustFail = _bot_module.SubmissionMustFail



class MustCopyError(Exception):
    pass

def _raise_must_copy(*args, **kwargs):
    raise MustCopyError(
        "Cannot modify a list that originated in Constants. "
        "First, you must make a copy of it, for example: "
        "'my_list = Constants.my_list.copy()' "
        "or "
        "'self.participant.vars['my_list'] = Constants.my_list.copy()'. "
        "This is to prevent accidentally modifying the original list. "
    )

class ConstantsList(list):

    __setitem__ = _raise_must_copy
    __delitem__ = _raise_must_copy
    clear = _raise_must_copy
    __iadd__ = _raise_must_copy
    __imul__ = _raise_must_copy
    append = _raise_must_copy
    extend = _raise_must_copy
    insert = _raise_must_copy
    pop = _raise_must_copy
    remove = _raise_must_copy
    reverse = _raise_must_copy
    sort = _raise_must_copy


class BaseConstantsMeta(type):
    def __setattr__(cls, attr, value):
        raise AttributeError("Constants are read-only.")

    def __new__(mcs, name, bases, attrs):

        for k, v in attrs.items():
            if type(v) == list:
                attrs[k] = ConstantsList(v)

        return super().__new__(mcs, name, bases, attrs)


class BaseConstants(metaclass=BaseConstantsMeta):
    pass



from django.core.management.base import BaseCommand
from glob import glob
import os.path
import sys

from contextlib import contextmanager
import io
import os

@contextmanager
def inplace_readonly(filename):
    readable = open(filename, 'r', encoding='utf-8')
    try:
        yield readable, None
    finally:
        readable.close()


@contextmanager
def inplace(filename, mode='r', buffering=-1, encoding='utf-8', errors=None,
            newline=None, backup_extension=None):
    # from Martijn Pieters blog
    """Allow for a file to be replaced with new content.

    yields a tuple of (readable, writable) file objects, where writable
    replaces readable.

    If an exception occurs, the old file is restored, removing the
    written data.

    mode should *not* use 'w', 'a' or '+'; only read-only-modes are supported.

    """

    # move existing file to backup, create new file with same permissions
    # borrowed extensively from the fileinput module
    if set(mode).intersection('wa+'):
        raise ValueError('Only read-only file modes can be used')

    backupfilename = filename + (backup_extension or os.extsep + 'bak')
    try:
        os.unlink(backupfilename)
    except os.error:
        pass
    os.rename(filename, backupfilename)
    readable = io.open(backupfilename, mode, buffering=buffering,
                       encoding=encoding, errors=errors, newline=newline)
    try:
        perm = os.fstat(readable.fileno()).st_mode
    except OSError:
        writable = open(filename, 'w' + mode.replace('r', ''),
                        buffering=buffering, encoding=encoding, errors=errors,
                        newline=newline)
    else:
        os_mode = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
        if hasattr(os, 'O_BINARY'):
            os_mode |= os.O_BINARY
        fd = os.open(filename, os_mode, perm)
        writable = io.open(fd, "w" + mode.replace('r', ''), buffering=buffering,
                           encoding=encoding, errors=errors, newline=newline)
        try:
            if hasattr(os, 'chmod'):
                os.chmod(filename, perm)
        except OSError:
            pass
    try:
        yield readable, writable
    except Exception:
        # move backup back
        try:
            os.unlink(filename)
        except os.error:
            pass
        os.rename(backupfilename, filename)
        raise
    finally:
        readable.close()
        writable.close()
        try:
            os.unlink(backupfilename)
        except os.error:
            pass

class Command(BaseCommand):
    help = "oTree: upgrade code to latest syntax."

    something_changed = False
    current_rule_number = 1

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        ahelp = (
            'Tells the command to NOT prompt the user for '
            'input of any kind.')
        parser.add_argument(
            '--noinput', action='store_false', dest='interactive',
            default=True, help=ahelp)

    def handle(self, *args, **options):

        self.stdout = sys.stdout
        self.dry_run = options["interactive"]

        if self.dry_run:
            self.scan()
            if self.something_changed:
                prompt = (
                    'The above changes will be made to your code. '
                    'You should back up your code before you proceed. '
                    'If you are using git, you should do a git commit before you proceed. '
                    'Proceed with syntax upgrade? (y or n): '
                )
                confirmed = input(prompt).lower() == 'y'
                if not confirmed:
                    sys.exit()
            else:
                self.stdout.write('Code is up to date; no changes to be made.')
                sys.exit()

        self.current_rule_number = 1
        self.dry_run = False
        self.scan()

    def scan(self):

        dry_run = self.dry_run

        python_fns = glob('*.py') + glob('*/*.py') + glob('*/*/*.py')
        for rule in PYTHON_RULES:
            self.print_rule_header()
            # don't use recursive, in case there is a venv.
            for fn in python_fns:
                self.apply_rule_to_file(fn, rule)

        html_fns = glob('**/*.html', recursive=True)
        for rule in HTML_RULES:
            self.print_rule_header()
            for fn in html_fns:
                self.apply_rule_to_file(fn, rule)

        test_fns = glob('tests.py') + glob('*/tests.py') + glob('*/*/tests.py')
        for rule in TESTS_RULES:
            self.print_rule_header()
            for fn in test_fns:
                self.apply_rule_to_file(fn, rule)

        if os.path.isfile('Procfile'):
            for rule in PROCFILE_RULES:
                self.print_rule_header()
                self.apply_rule_to_file('Procfile', rule)

        self.print_rule_header()
        view_fns = glob('views.py') + glob('*/views.py') + glob('*/*/views.py')
        for fn in view_fns:
            new_fn = fn.replace('views.py', 'pages.py')
            if dry_run:
                self.print_change('Rename {} to {}\n'.format(fn, new_fn))
            else:
                os.rename(fn, new_fn)

        self.print_rule_header()
        for fn in ['requirements_base.txt', 'requirements.txt', 'requirements_heroku.txt']:
            if os.path.isfile(fn):
                if dry_run:
                    with open(fn, 'r') as infh:
                        for line in infh:
                            if line.startswith('Django=='):
                                self.print_change(fn + '\n')
                                self.print_change('REMOVE: {}'.format(line))
                else:
                    with inplace(fn, 'r') as (infh, outfh):
                        for line in infh:
                            if line.startswith('otree-core'):
                                outfh.write('otree\n')
                            else:
                                l_django = line.startswith('Django==')
                                # remove obsolete comments about including Django
                                l_explicitly = 'explicitly' in line
                                l_collectstatic = 'collectstatic' in line
                                if not (l_django or l_collectstatic or l_explicitly):
                                    outfh.write(line)

        self.print_rule_header()
        if os.path.isfile('.gitignore'):
            temp_wildcard = '__temp*'
            with open('.gitignore', 'r') as file:
                temp_wildcard_found = temp_wildcard in file.read()
            if not temp_wildcard_found:
                if dry_run:
                    self.print_change('ADD "__temp*" to .gitignore\n')
                else:
                    with open('.gitignore', 'a') as file:
                        file.write('\n' + temp_wildcard)

        self.print_rule_header()
        for fn in ['_templates/global/Base.html', '_templates/global/Page.html']:
            rule = Rule(
                '{% extends "otree/FormPage.html" %}',
                '{% extends "otree/Page.html" %}',
            )
            if os.path.isfile(fn):
                self.apply_rule_to_file(fn, rule)

    def apply_rule_to_file(self, fn, rule):
        if self.dry_run:
            cm = inplace_readonly(fn)
        else:
            cm = inplace(fn, 'r')
        with cm as (infh, outfh):
            lines = list(infh)
            for i, line in enumerate(lines):
                new_line = rule.new_line(line)
                if new_line != line:
                    if self.dry_run:
                        self.print_change('{}, line {}\n'.format(fn, i + 1))
                        self.print_change('BEFORE: {}'.format(line))
                        self.print_change('AFTER:  {}'.format(new_line))
                if not self.dry_run:
                    outfh.write(new_line)

    def print_rule_header(self):
        self.stdout.write('****APPLYING RULE #{}\n'.format(self.current_rule_number))
        self.current_rule_number += 1

    def print_change(self, msg):
        self.something_changed = True
        self.stdout.write(msg)

    # TODO: remove migrations?

class Rule:
    def __init__(self, old_token, new_token, description=None):
        self.old_token = old_token
        self.new_token = new_token
        self.description = description or 'rename {} to {}'.format(old_token, new_token)

    def new_line(self, line):
        return line.replace(self.old_token, self.new_token)


PYTHON_RULES = [
    Rule(
        'form_model = models.Player',
        "form_model = 'player'",
        "change <form_model = models.Player> to <form_model = 'player'>"
    ),
    # if they're using Django models, this won't work.
    Rule(
        '= models.CharField',
        '= models.StringField',
    ), # what about initial=None and default=None? maybe do a simple lookahead
    # if they're using Django models, this won't work.
    Rule(
        '= models.TextField',
        '= models.LongStringField',
    ), # what about initial=None and default=None?
    Rule(
        'form_model = models.Group',
        "form_model = 'group'",
        "change <form_model = models.Player> to <form_model = 'group'>"
    ),
    Rule(
        'before_session_starts',
        'creating_session',
    ),
    # if they are using Django models, or if there are migrations, this won't work.
    # Rule(
    #     'verbose_name',
    #    'label',
    # ),
    Rule(
        'SliderInput',
        'Slider',
    )
]

TESTS_RULES = [
    Rule(
        'views',
        'pages',
        'fix references to views.py'
    ),
]

HTML_RULES = [
    Rule(
        '{% load otreechat %}',
        ''
    ),
    Rule(
        'with label=',
        'label=',
        'remove unnecessary "with" in formfield tag'
    ),
    Rule(
        'otree_tags',
        'otree',
    ),
    # someone might have created their own next button;
    # we don't want to break the selector for that.
    #Rule(
    #    '.next-button',
    #    '.otree-btn-next'
    #),
    Rule(
        '.otree-next-button',
        '.otree-btn-next'
    ),
    Rule(
        '#otree-body',
        '.otree-body'
    ),
    Rule(
        '#otree-title',
        '.otree-title'
    ),
    Rule(
        '#otree-timeout',
        '.otree-timer'
    ),
    Rule(
        '#otree-form-errors',
        '.otree-form-errors'
    ),
    # rarely used rules...will just slow things down
    #Rule(
    #    '#otree-wait-page-body',
    #    '.otree-wait-page'
    #),
    #Rule(
    #    '#otree-wait-page-title-text',
    #    '.otree-wait-page__title'
    #),
    #Rule(
    #    '#otree-wait-page-body-text',
    #    '.otree-wait-page__body'
    #),
]


PROCFILE_RULES = [
    Rule(
        'web: otree webandworkers',
        'web: otree runprodserver1of2',
    ),
    Rule(
        'timeoutworker: otree timeoutworker',
        'worker: otree runprodserver2of2',
    )
]




from django.core.management.base import BaseCommand, CommandError
import tarfile
import logging
import os.path
import sys
from pathlib import Path
from django.conf import settings

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Unzip a zipped oTree project"

    def add_arguments(self, parser):
        parser.add_argument(
            'zip_file', type=str, help="The .otreezip file")
        # it's good to require this arg because then it's obvious that the files
        # will be put in that subfolder, and not dumped in the current dir
        parser.add_argument(
            'output_folder', type=str, nargs='?',
            help="What to call the new project folder")

    def handle(self, **options):
        zip_file = options['zip_file']
        output_folder = options['output_folder'] or auto_named_output_folder(zip_file)
        unzip(zip_file, output_folder)
        msg = (
            f'Unzipped file. Enter this:\n'
            f'cd {esc_fn(output_folder)}\n'
            "then run 'pip3 install -r requirements.txt' to install this project's dependencies."
        )

        logger.info(msg)


    def run_from_argv(self, argv):
        '''
        override this because the built-in django one executes system checks,
        which trips because settings are not configured.
        as at 2018-11-19, 'unzip' is the only
        otree-specific management command that doesn't require settings
        '''

        if len(argv) == 2:
            self.stdout.write(
                'You must provide the name of the *.otreezip file. Example:\n '
                'otree unzip AAA.otreezip'
            )
            sys.exit(-1)

        parser = self.create_parser(argv[0], argv[1])
        options = parser.parse_args(argv[2:])
        cmd_options = vars(options)
        self.handle(**cmd_options)


def esc_fn(fn):
    if ' ' in fn:
        return f'\"{fn}\"'
    return fn

def auto_named_output_folder(zip_file_name) -> str:
    default_folder_name = Path(zip_file_name).stem

    if not Path(default_folder_name).exists():
        return default_folder_name

    logger.info(
        'Hint: you can provide the name of the folder to create. Example:\n'
        f"otree unzip {esc_fn(zip_file_name)} my_project"
    )
    for x in range(2, 20):
        folder_name = f'{default_folder_name}-{x}'
        if not Path(folder_name).exists():
            return folder_name
    logger.error(
        f"Could not unzip the file; target folder {folder_name} already exists. "
    )
    sys.exit(-1)


def unzip(zip_file: str, output_folder):
    if os.path.isfile('settings.py') and os.path.isfile('manage.py'):
        logger.error(
            'You are trying to create a project but it seems you are '
            'already in a project folder (found settings.py and manage.py).'
        )
        sys.exit(-1)

    with tarfile.open(zip_file) as tar:
        tar.extractall(output_folder)




from django.core.signing import Signer


def create_session_group_name(pre_create_id):
    return 'wait_for_session_{}'.format(pre_create_id)


def wait_page_group_name(session_id, page_index,
                         group_id_in_subsession=''):

    return 'wait-page-{}-page{}-{}'.format(
        session_id, page_index, group_id_in_subsession)


def gbat_group_name(session_id, page_index):
    return 'group_by_arrival_time_session{}_page{}'.format(
        session_id, page_index)

def gbat_path(session_id, index_in_pages, app_name, player_id):
    return '/group_by_arrival_time/{},{},{},{}/'.format(
        session_id, index_in_pages, app_name, player_id
        )

def room_participants_group_name(room_name):
    return 'room-participants-{}'.format(room_name)


def room_participant_path(room_name, participant_label, tab_unique_id):
    return '/wait_for_session_in_room/{},{},{}/'.format(
            room_name, participant_label, tab_unique_id
    )

def room_admin_path(room_name):
    return '/room_without_session/{}/'.format(room_name)

def wait_for_session_path(pre_create_id):
    return '/wait_for_session/{}/'.format(pre_create_id)

def wait_page_path(session_pk, index_in_pages, group_id_in_subsession=''):
    return '/wait_page/{},{},{}/'.format(
        session_pk, index_in_pages, group_id_in_subsession
    )

def browser_bots_launcher_group(session_code):
    return 'browser-bots-client-{}'.format(session_code)

def browser_bots_launcher_path(session_code):
    return '/browser_bots_client/{}/'.format(session_code)

def auto_advance_path(participant_code, page_index):
    return '/auto_advance/{},{}/'.format(participant_code, page_index)


def chat_path(channel, participant_id):
    channel_and_id = '{}/{}'.format(channel, participant_id)
    channel_and_id_signed = Signer(sep='/').sign(channel_and_id)

    return '/otreechat_core/{}/'.format(channel_and_id_signed)


def get_chat_group(channel):
    return 'otreechat-{}'.format(channel)


#!/usr/bin/env python
# -*- coding: utf-8 -*-


# =============================================================================
# IMPORTS
# =============================================================================

from django.core.management.base import BaseCommand
import sys
from otree.models import Session


# =============================================================================
# COMMND
# =============================================================================

class Command(BaseCommand):
    help = ("List all available sessions")

    def handle(self, **options):
        rows = []
        for session in Session.objects.all():
            sconfig = session.config
            rows.append({
                "name": sconfig["display_name"],
                "code": session.code,
                "participants": str(session.participant_set.count()),
                "appsequence": ", ".join(sconfig["app_sequence"])
            })

        if rows:
            rows.insert(0, {
                "name": "Name", "code": "Code",
                "participants": "Participants", "appsequence": "App Sequence"})

            name_just = max(len(r["name"]) for r in rows) + 2
            code_just = max(len(r["code"]) for r in rows) + 2
            participants_just = max(len(r["participants"]) for r in rows) + 2
            appsequence_just = max(len(r["appsequence"]) for r in rows) + 2

            for idx, row in enumerate(rows):
                if idx == 1:
                    sys.stdout.write(
                        "-" * name_just, "-" * code_just,
                        "-" * participants_just, "-" * appsequence_just)
                self.stdout.write(
                    row["name"].ljust(name_just),
                    row["code"].ljust(code_just),
                    row["participants"].ljust(participants_just),
                    row["appsequence"].ljust(appsequence_just))
        self.stdout.write('\n')



# these symbols are a fallback if we don't have an explicit rule
# for the currency/language combination.
# (most common situation is where the language is English)

CURRENCY_SYMBOLS = {
    'AED': 'AED',
    'ARS': '$',
    'AUD': '$',
    'BRL': 'R$',
    'CAD': '$',
    'CHF': 'CHF',
    # 2017-11-29: this used to be ￥, don't know where that came from
    'CNY': '¥',
    'CZK': 'Kč',
    'DKK': 'kr',
    'EGP': 'ج.م.‏',
    'EUR': '€',
    'GBP': '£',
    'HKD': 'HK$',
    'HUF': 'Ft',
    'ILS': '₪',
    'INR': '₹',
    'JPY': '¥',
    'KRW': '₩',
    'MXN': '$',
    'MYR': 'RM',
    'NOK': 'kr',
    'PLN': 'zł',
    'RUB': '₽',
    'SEK': 'kr',
    'SGD': 'SGD',
    'THB': 'THB',
    'TRY': '₺',
    'TWD': '$',
    'USD': '$',
    'ZAR': 'R',
}


def get_currency_format(lc: str, LO: str, CUR: str) -> str:

    '''because of all the if statements, this has very low code coverage
    but it's ok'''

    ##############################
    # Languages with complex rules
    ##############################

    if lc == 'en':
        if CUR in ['USD', 'CAD', 'AUD']:
            return '$#'
        if CUR == 'GBP':
            return '£#'
        if CUR == 'EUR':
            return '€#'
        if CUR == 'INR':
            return '₹ #'
        if CUR == 'SGD':
            return '$#'
        return '¤#'

    if lc == 'zh':
        if CUR == 'CNY':
            return '¥#'
        if CUR == 'HKD':
            return 'HK$#'
        if CUR == 'TWD':
            return '$#'
        if CUR == 'SGD':
            return 'SGD#'
        return '¤#'

    if lc == 'de':
        if CUR == 'EUR':
            if LO == 'AT':
                return '€ #'
            return '# €'
        if CUR == 'CHF':
            return 'CHF #'
        return '¤ #'

    if lc == 'es':
        if CUR == 'ARS':
            return '$ #'
        if CUR == 'EUR':
            return '# €'
        if CUR == 'MXN':
            return '$#'
        return '# ¤'

    if lc == 'nl':
        if LO == 'BE':
            if CUR == 'EUR':
                return '# €'
            return '# ¤'
        # default NL
        if CUR == 'EUR':
            return '€ #'
        return '¤ #'

    if lc == 'pt':
        if CUR == 'BRL':
            return 'R$#'
        if CUR == 'EUR':
            return '# €'
        return '¤#'

    if lc == 'ar':
        if CUR == 'AED':
            return 'د.إ.‏ #'
        return '¤ #'

    #############################
    # Languages with simple rules
    #############################

    if lc == 'cs':
        if CUR == 'CZK':
            return '# Kč'
        return '# ¤'
    if lc == 'da':
        if CUR == 'DKK':
            return '# kr.'
        return '# ¤'
    if lc == 'fi':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'fr':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'he':
        if CUR == 'ILS':
            return '# ₪'
        return '# ¤'
    if lc == 'hu':
        if CUR == 'HUF':
            return '# Ft'
        return '# ¤'
    if lc == 'it':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'ja':
        if CUR == 'JPY':
            return '¥#'
        return '¤#'
    if lc == 'ko':
        if CUR == 'KRW':
            return '₩#'
        return '¤#'
    if lc == 'ms':
        if CUR == 'MYR':
            return 'RM#'
        return '¤#'
    if lc == 'nb':
        if CUR == 'NOK':
            return 'kr #'
        return '¤ #'
    if lc == 'pl':
        if CUR == 'PLN':
            return '# zł'
        return '# ¤'
    if lc == 'ru':
        if CUR == 'RUB':
            return '# ₽'
        return '# ¤'
    if lc == 'sv':
        if CUR == 'SEK':
            return '# kr'
        return '# ¤'
    if lc == 'th':
        if CUR == 'THB':
            return 'THB#'
        return '¤#'
    if lc == 'tr':
        if CUR == 'TRY':
            return '# ₺'
        return '# ¤'
    if lc == 'zu':
        if CUR == 'ZAR':
            return 'R#'
        return '¤#'

    # fallback if it's another language, etc.
    return '# ¤'


'''
This file was built by taking common currencies and reverse engineering
the rules using Babel:

curs = [('ar_AE', 'AED'),
 ('ar_EG', 'EGP'),
 ('cmn', 'SGD'),
 ('cs', 'CZK'),
 ('da_DK', 'DKK'),
 ('de', 'EUR'),
 ('de_AT', 'EUR'),
 ('de_CH', 'CHF'),
 ('en_AU', 'AUD'),
 ('en_CA', 'CAD'),
 ('en_GB', 'GBP'),
 ('en_IE', 'EUR'),
 ('en_IN', 'INR'),
 ('en_US', 'USD'),
 ('es_AR', 'ARS'),
 ('es_ES', 'EUR'),
 ('es_MX', 'MXN'),
 ('fi_FI', 'EUR'),
 ('fr_FR', 'EUR'),
 ('he', 'ILS'),
 ('hu_HU', 'HUF'),
 ('it_IT', 'EUR'),
 ('ja', 'JPY'),
 ('ko_KR', 'KRW'),
 ('ms_MY', 'MYR'),
 ('nl_BE', 'EUR'),
 ('nl_NL', 'EUR'),
 ('no', 'NOK'),
 ('pl', 'PLN'),
 ('pt_BR', 'BRL'),
 ('pt_PT', 'EUR'),
 ('ru', 'RUB'),
 ('sv_SE', 'SEK'),
 ('th', 'THB'),
 ('tr_TR', 'TRY'),
 ('zh_CN', 'CNY'),
 ('zh_HK', 'HKD'),
 ('zh_TW', 'TWD'),
 ('zu', 'ZAR')]

for lc, CUR in curs:
    if '_' in lc:
        la, LO = lc.split('_')
    else:
        la, LO = lc, ''
    locale = Locale.parse(lc)
    pattern = locale.currency_formats['standard'].pattern.replace('\\xa4', ' ').replace('#,##0.00', '#')    
    formatted = fc(1.00, currency=CUR, locale=lc)
    print('{}\t{}\t{}\t{}'.format(la, LO, CUR, formatted))

ar	AE	AED	د.إ.‏ 1.00
ar	EG	EGP	ج.م.‏ 1.00
cmn		SGD	SGD1.00
cs		CZK	1,00 Kč
da	DK	DKK	1,00 kr.
de		EUR	1,00 €
de	AT	EUR	€ 1,00
de	CH	CHF	CHF 1.00
en	AU	AUD	$1.00
en	CA	CAD	$1.00
en	GB	GBP	£1.00
en	IE	EUR	€1.00
en	IN	INR	₹ 1.00
en	US	USD	$1.00
es	AR	ARS	$ 1,00
es	ES	EUR	1,00 €
es	MX	MXN	$1.00
fi	FI	EUR	1,00 €
fr	FR	EUR	1,00 €
he		ILS	1.00 ₪
hu	HU	HUF	1,00 Ft
it	IT	EUR	1,00 €
ja		JPY	￥1
ko	KR	KRW	₩1
ms	MY	MYR	RM1.00
nl	BE	EUR	1,00 €
nl	NL	EUR	€ 1,00
no		NOK	kr 1,00
pl		PLN	1,00 zł
pt	BR	BRL	R$1,00
pt	PT	EUR	1,00 €
ru		RUB	1,00 ₽
sv	SE	SEK	1,00 kr
th		THB	THB1.00
tr	TR	TRY	1,00 ₺
zh	CN	CNY	￥1.00
zh	HK	HKD	HK$1.00
zh	TW	TWD	$1.00
zu		ZAR	R1.00

Almost all currencies just put the negative sign in front:

ar	AE	AED -د.إ.‏ 1.00
ar	EG	EGP -ج.م.‏ 1.00
cmn		SGD	-SGD1.00
cs		CZK	-1,00 Kč
da	DK	DKK	-1,00 kr.
de		EUR	-1,00 €
de	AT	EUR	-€ 1,00
de	CH	CHF	CHF-1.00
en	AU	AUD	-$1.00
en	CA	CAD	-$1.00
en	GB	GBP	-£1.00
en	IE	EUR	-€1.00
en	IN	INR	-₹ 1.00
en	US	USD	-$1.00
es	AR	ARS	-$ 1,00
es	ES	EUR	-1,00 €
es	MX	MXN	-$1.00
fi	FI	EUR	-1,00 €
fr	FR	EUR	-1,00 €
he		ILS	-1.00 ₪
hu	HU	HUF	-1,00 Ft
it	IT	EUR	-1,00 €
ja		JPY	-￥1
ko	KR	KRW	-₩1
ms	MY	MYR	-RM1.00
nl	BE	EUR	-1,00 €
nl	NL	EUR	€ -1,00
no		NOK	-kr 1,00
pl		PLN	-1,00 zł
pt	BR	BRL	-R$1,00
pt	PT	EUR	-1,00 €
ru		RUB	-1,00 ₽
sv	SE	SEK	-1,00 kr
th		THB	-THB1.00
tr	TR	TRY	-1,00 ₺
zh	CN	CNY	-￥1.00
zh	HK	HKD	-HK$1.00
zh	TW	TWD	-$1.00
zu		ZAR	-R1.00

In English:

# comment: this is maybe not optimal for our use, 
because experiments in non-USD currencies are usually run locally, 
where people are already familiar with the currency. 
In Canada, people don't write 'CA$1.00', they just write $1.00.

AED en_US AED1.00
EGP en_US E£1.00
SGD en_US $1.00
CZK en_US Kč1.00
DKK en_US kr1.00
EUR en_US €1.00
EUR en_US €1.00
CHF en_US CHF1.00
AUD en_US A$1.00
CAD en_US CA$1.00
GBP en_US £1.00
EUR en_US €1.00
INR en_US ₹1.00
USD en_US $1.00
ARS en_US $1.00
EUR en_US €1.00
MXN en_US MX$1.00
EUR en_US €1.00
EUR en_US €1.00
ILS en_US ₪1.00
HUF en_US Ft1.00
EUR en_US €1.00
JPY en_US ¥1
KRW en_US ₩1
MYR en_US RM1.00
EUR en_US €1.00
EUR en_US €1.00
NOK en_US kr1.00
PLN en_US zł1.00
BRL en_US R$1.00
EUR en_US €1.00
RUB en_US ₽1.00
SEK en_US kr1.00
THB en_US ฿1.00
TRY en_US ₺1.00
CNY en_US CN¥1.00
HKD en_US HK$1.00
TWD en_US NT$1.00
ZAR en_US R1.00

'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# flake8: noqa

from django.forms import *

# Delete `fields` and `widgets` attributes which came from Django and
# would otherwise make it impossible to import e.g. the otree.forms.fields
# module.
del fields
del widgets

from .fields import *
from .forms import *
from .widgets import *



import os
import os.path
from django.contrib.messages import constants as messages
import dj_database_url

DEFAULT_MIDDLEWARE = (
    'otree.middleware.CheckDBMiddleware',
    'otree.middleware.perf_middleware',
    'whitenoise.middleware.WhiteNoiseMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',


    # 2015-04-08: disabling SSLify until we make this work better
    # 'sslify.middleware.SSLifyMiddleware',
)


def collapse_to_unique_list(*args):
    """Create a new list with all elements from a given lists without reapeated
    elements

    """
    combined = []
    for arg in args:
        for elem in arg or ():
            if elem not in combined:
                combined.append(elem)
    return combined


def get_default_settings(user_settings: dict):
    '''
    doesn't mutate user_settings, just reads from it
    because some settings depend on others
    '''
    default_settings = {}

    sentry_dsn = user_settings.get('SENTRY_DSN') or os.environ.get('SENTRY_DSN')
    if sentry_dsn:
        default_settings['RAVEN_CONFIG'] = {
            'dsn': sentry_dsn,
            'processors': ['raven.processors.SanitizePasswordsProcessor'],
            # 2018-11-24: breadcrumbs were causing memory leaks when doing queries,
            # especially when creating sessions, which construct hugely verbose
            # queries with bulk_create.
            # however, i could only clearly observe the difference this line makes
            # when testing
            # on a script that bulk_created thousands of non-otree models.
            # when testing on a live server, things are more ambiguous.
            # maybe just refreshing the page several times after creating a session
            # is enough to reset memory to reasnoable levels?
            # disabling also may make things faster...
            # in anecdotal test, 40 vs 50 seconds
            'enable_breadcrumbs': False,
        }
        # SentryHandler is very slow with URL resolving...can add 2 seconds
        # to runserver startup! so only use when it's needed
        sentry_handler_class = 'raven.contrib.django.raven_compat.handlers.SentryHandler'
    else:
        sentry_handler_class = 'logging.StreamHandler'

    logging = {
        'version': 1,
        'disable_existing_loggers': False,
        'root': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'formatters': {
            'verbose': {
                'format': '[%(levelname)s|%(asctime)s] %(name)s > %(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'INFO',
                'class': 'logging.StreamHandler',
                'formatter': 'simple'
            },
            'sentry': {
                'level': 'WARNING',
                # perf issue. just use StreamHandler by default.
                # only use the real sentry handler if a DSN exists.
                # see below.
                'class': sentry_handler_class,

            },
        },
        'loggers': {
            'otree.test.core': {
                'handlers': ['console'],
                'propagate': False,
                'level': 'INFO',
            },
            # 2016-07-25: botworker seems to be sending messages to Sentry
            # without any special configuration, not sure why.
            # but, i should use a logger, because i need to catch exceptions
            # in botworker so it keeps running
            'otree.test.browser_bots': {
                'handlers': ['sentry', 'console'],
                'propagate': False,
                'level': 'INFO',
            },
            'django.request': {
                'handlers': ['console'],
                'propagate': True,
                'level': 'DEBUG',
            },
            # logger so that we can explicitly send certain warnings to sentry,
            # without raising an exception.
            # 2016-10-23: has not been used yet
            'otree.sentry': {
                'handlers': ['sentry'],
                'propagate': True,
                'level': 'DEBUG',
            },
            # log any error that occurs inside channels code
            'django.channels': {
                'handlers': ['sentry'],
                'propagate': True,
                'level': 'ERROR',
            },
            # This is required for exceptions inside Huey tasks to get logged
            # to Sentry
            'huey.consumer': {
                'handlers': ['sentry', 'console'],
                'level': 'INFO'
            },
            # suppress the INFO message: 'raven is not configured (logging
            # disabled).....', in case someone doesn't have a DSN
            'raven.contrib.django.client.DjangoClient': {
                'handlers': ['console'],
                'level': 'WARNING'
            }


        }
    }

    REDIS_URL = os.environ.get('REDIS_URL', 'redis://localhost:6379')
    BASE_DIR = user_settings.get('BASE_DIR', '')

    default_settings.update({
        'DEBUG': os.environ.get('OTREE_PRODUCTION') in [None, '', '0'],
        'AWS_ACCESS_KEY_ID': os.environ.get('AWS_ACCESS_KEY_ID'),
        'AWS_SECRET_ACCESS_KEY': os.environ.get('AWS_SECRET_ACCESS_KEY'),
        'AUTH_LEVEL': os.environ.get('OTREE_AUTH_LEVEL'),
        'DATABASES': {
            'default': dj_database_url.config(
                default='sqlite:///' + os.path.join(BASE_DIR, 'db.sqlite3')
            )
        },
        'HUEY': {
            'name': 'otree-huey',
            'connection': {
                'url': REDIS_URL,
            },
            'always_eager': False,
            # I need a result store to retrieve the results of browser-bots
            # tasks and pinging, even if the result is evaluated immediately
            # (otherwise, calling the task returns None.
            'result_store': False,
            'consumer': {
                'workers': 1,
                # 'worker_type': 'thread',
                'scheduler_interval': 5,
                'loglevel': 'warning',
            },
        },
        # set to True so that if there is an error in an {% include %}'d
        # template, it doesn't just fail silently. instead should raise
        # an error (and send through Sentry etc)
        'STATIC_ROOT': os.path.join(BASE_DIR, '__temp_static_root'),
        'STATIC_URL': '/static/',
        'STATICFILES_STORAGE': (
            'whitenoise.django.GzipManifestStaticFilesStorage'
        ),
        'ROOT_URLCONF': 'otree.urls',

        'TIME_ZONE': 'UTC',
        'USE_TZ': True,
        'ALLOWED_HOSTS': ['*'],

        'LOGGING': logging,

        'FORM_RENDERER':  'django.forms.renderers.TemplatesSetting',

        'REAL_WORLD_CURRENCY_CODE': 'USD',
        'REAL_WORLD_CURRENCY_DECIMAL_PLACES': 2,
        'USE_POINTS': True,
        'POINTS_DECIMAL_PLACES': 0,

        # INSTALLED_APPS is required somehow to see if settings.configured()
        # alse we should encourage it to be defined in project, because it helps
        # pycharm autocomplete

        #'INSTALLED_APPS': ['otree'],

        'ADMIN_PASSWORD': os.environ.get('OTREE_ADMIN_PASSWORD', ''),
        # eventually can remove this,
        # when it's present in otree-library
        # that most people downloaded
        'USE_L10N': True,
        'SECURE_PROXY_SSL_HEADER': ('HTTP_X_FORWARDED_PROTO', 'https'),

        # The project can override the routing.py used as entry point by
        # setting CHANNEL_ROUTING.

        'CHANNEL_LAYERS': {
            'default': {
                "BACKEND": "otree.channels.asgi_redis.RedisChannelLayer",
                "CONFIG": {
                    "hosts": [REDIS_URL],
                },
                'ROUTING': user_settings.get(
                    'CHANNEL_ROUTING',
                    'otree.channels.routing.channel_routing'),
            },
            # note: if I start using ChannelsLiveServerTestCase again,
            # i might have to move this out,
            # but because it doesn't work with multiple
            # channel layers.
            'inmemory': {
                "BACKEND": "asgiref.inmemory.ChannelLayer",
                'ROUTING': user_settings.get(
                    'CHANNEL_ROUTING',
                    'otree.channels.routing.channel_routing'),
            },
        },

        # for convenience within oTree
        'REDIS_URL': REDIS_URL,

        # since workers on Amazon MTurk can return the hit
        # we need extra participants created on the
        # server.
        # The following setting is ratio:
        # num_participants_server / num_participants_mturk
        'MTURK_NUM_PARTICIPANTS_MULTIPLE': 2,
        'LOCALE_PATHS': [
            os.path.join(user_settings.get('BASE_DIR', ''), 'locale')
        ],

        # ideally this would be a per-app setting, but I don't want to
        # pollute Constants. It doesn't make as much sense per session config,
        # so I'm just going the simple route and making it a global setting.
        'BOTS_CHECK_HTML': True,
    })
    return default_settings



class InvalidVariableError(Exception):
    pass


class InvalidTemplateVariable(str):
    def get_error_message(self, variable_name_dotted: str):
        bits = variable_name_dotted.split('.')
        if len(bits) == 1:
            return (
                'Invalid variable: "{}". '
                'Maybe you need to return it from vars_for_template()'
            ).format(bits[0])

        built_in_vars = [
            'player',
            'group',
            'subsession',
            'participant',
            'session',
            'Constants',
        ]

        if bits[0] in built_in_vars:
            # This will not make sense in the admin report!
            # but that's OK, it's a rare case, more advanced users
            return (
                '{} has no attribute "{}"'
            ).format(bits[0], '.'.join(bits[1:]))
        elif bits[0] == 'self' and bits[1] in built_in_vars:
            return (
                "Don't use 'self' in the template. "
                "Just write: {}"
            ).format('.'.join(bits[1:]))
        else:
            return 'Invalid variable: {}'.format(variable_name_dotted)

    def __mod__(self, other):
        '''hack that takes advantage of string_if_invalid's %s behavior'''
        msg = self.get_error_message(str(other))
        # "from None" because otherwise we get the full chain of
        # checking if it's an attribute, dict key, list index ...
        raise InvalidVariableError(msg) from None


def validate_user_settings(settings: dict):
    '''
    These are the settings that from my testing must not be None.
    This just exists so that oTree doesn't give a cryptic error later on
    '''

    # currently not using the datatypes, maybe do that later
    required_settings = {
        'SESSION_CONFIG_DEFAULTS': dict,
        'SESSION_CONFIGS': list,
        'LANGUAGE_CODE': str,
        'SECRET_KEY': str,
        'ADMIN_USERNAME': str,
    }
    for SETTING in required_settings:
        if SETTING not in settings:
            raise ValueError(f'Required setting {SETTING} is missing from settings.py.')
        elif settings.get(SETTING) is None:
            raise ValueError(f'settings.py: setting {SETTING} cannot be None.')


def augment_settings(settings: dict):
    validate_user_settings(settings)
    default_settings = get_default_settings(settings)
    for k, v in default_settings.items():
        settings.setdefault(k, v)

    all_otree_apps_set = set()

    for s in settings['SESSION_CONFIGS']:
        for app in s['app_sequence']:
            all_otree_apps_set.add(app)

    all_otree_apps = list(all_otree_apps_set)

    no_experiment_apps = [
        'otree',

        # django.contrib.auth is slow, about 300ms.
        # would be nice to only add it if there is actually a password
        # i tried that but would need to add various complicated "if"s
        # throughout the code
        'django.contrib.auth',
        'django.forms',
        # needed for auth and very quick to load
        'django.contrib.contenttypes',
        'django.contrib.sessions',
        'django.contrib.messages',
        # need to keep this around indefinitely for all the people who
        # have {% load staticfiles %}
        'django.contrib.staticfiles',
        'channels',
        'huey.contrib.djhuey',
        'idmap',
    ]

    # these are slow...only add if we need them
    if settings.get('RAVEN_CONFIG'):
        no_experiment_apps.append('raven.contrib.django.raven_compat')

    if os.environ.get('OTREE_SECRET_KEY'):
        # then override the SECRET_KEY from settings file, which might
        # be exposed if the source code is made public
        # an alternative is to change the project template so that
        # SECRET_KEY = os.environ.get('OTREE_SECRET_KEY', '{{ secret_key }}')
        # but this change is just as good and backward compatible
        # and doesn't involve any churn.
        # people don't usually care about the specific value of their secret key.
        # eventually maybe we can make SECRET_KEY an optional setting,
        # but that will require people to set the env vars.
        settings['SECRET_KEY'] = os.environ['OTREE_SECRET_KEY']

    # order is important:
    # otree unregisters User & Group, which are installed by auth.
    # otree templates need to get loaded before the admin.
    no_experiment_apps = collapse_to_unique_list(
        no_experiment_apps,
        settings['INSTALLED_APPS'],
        settings.get('EXTENSION_APPS', [])
    )

    new_installed_apps = collapse_to_unique_list(
        no_experiment_apps, all_otree_apps)

    # TEMPLATES
    _template_dir = os.path.join(settings['BASE_DIR'], '_templates')
    if os.path.exists(_template_dir):
        new_template_dirs = [_template_dir]
    else:
        new_template_dirs = []

    # STATICFILES
    _static_dir = os.path.join(settings['BASE_DIR'], '_static')

    if os.path.exists(_static_dir):
        additional_static_dirs = [_static_dir]
    else:
        additional_static_dirs = []

    new_staticfiles_dirs = collapse_to_unique_list(
        settings.get('STATICFILES_DIRS'),
        additional_static_dirs,
    )

    new_middleware = collapse_to_unique_list(
        DEFAULT_MIDDLEWARE,
        settings.get('MIDDLEWARE_CLASSES'))

    augmented_settings = {
        'INSTALLED_APPS': new_installed_apps,
        'TEMPLATES': [{
            'BACKEND': 'django.template.backends.django.DjangoTemplates',
            'DIRS': new_template_dirs,
            'OPTIONS': {
                # 2016-10-08: setting template debug back to True,
                # because if an included template has an error, we need
                # to surface the error, rather than not showing the template.
                # that's how I set it in d1cd00ebfd43c7eff408dea6363fd14bb90e7c06,
                # but then in 2c10188b33f2ac36c046f4f0f8764e15d6a6fa81,
                # i set this to False, but I'm not sure why and there is no
                # note in the commit explaining why.
                'debug': True,
                'string_if_invalid': InvalidTemplateVariable("%s"),

                # in Django 1.11, the cached template loader is applied
                # automatically if template 'debug' is False,
                # but for now we need 'debug' True because otherwise
                # {% include %} fails silently.
                # in django 2.1, we can remove:
                # - the explicit 'debug': True
                # - 'loaders' below
                # - the patch in runserver.py
                # as long as we set 'APP_DIRS': True
                'loaders': [
                    ('django.template.loaders.cached.Loader', [
                        'django.template.loaders.filesystem.Loader',
                        'django.template.loaders.app_directories.Loader',
                    ]),
                ],
                'context_processors': (
                    # default ones in Django 1.8
                    'django.contrib.auth.context_processors.auth',
                    'django.template.context_processors.media',
                    'django.template.context_processors.static',
                    'django.contrib.messages.context_processors.messages',
                    'django.template.context_processors.request',
                 )
            },
        }],
        'STATICFILES_DIRS': new_staticfiles_dirs,
        'MIDDLEWARE': new_middleware,
        'INSTALLED_OTREE_APPS': all_otree_apps,
        'MESSAGE_TAGS': {messages.ERROR: 'danger'},
        'LOGIN_REDIRECT_URL': 'Sessions',
    }

    settings.update(augmented_settings)



from collections import namedtuple
from django.template.base import TextNode
from django.template.library import InclusionNode
from django.template.loader_tags import ExtendsNode, BlockNode
from django.utils.encoding import force_text
from itertools import chain
import unicodedata
import io

from otree.templatetags.otree import NEXT_BUTTON_TEMPLATE_PATH

class TemplateCheckContent(object):
    def __init__(self, root):
        self.root = root

    def node_is_empty(self, node):
        if isinstance(node, TextNode):
            return node.s.isspace()
        return False

    def is_extending(self, root):
        return any(
            isinstance(node, ExtendsNode)
            for node in root.nodelist)

    def is_content_node(self, node):
        """
        Returns if the node is an unempty text node.
        """
        if isinstance(node, TextNode):
            return not self.node_is_empty(node)
        return False

    def get_toplevel_content_nodes(self, root):
        nodes = []
        for node in root.nodelist:
            if isinstance(node, ExtendsNode):
                new_child_nodes = self.get_toplevel_content_nodes(node)
                nodes.extend(new_child_nodes)
            if self.is_content_node(node):
                nodes.append(node)
        return nodes

    def get_unreachable_content(self):
        """
        Return all top-level text nodes when the template is extending another
        template. Those text nodes won't be displayed during rendering since
        only content inside of blocks is considered in inheritance.
        """
        if not self.is_extending(self.root):
            return []

        textnodes = self.get_toplevel_content_nodes(self.root)
        return [node.s for node in textnodes]


def get_unreachable_content(root):
    check = TemplateCheckContent(root)
    return check.get_unreachable_content()


class TemplateCheckNextButton(object):
    def __init__(self, root):
        self.root = root

    def get_next_button_nodes(self, root):
        nodes = []
        for node in root.nodelist:
            if isinstance(node, (ExtendsNode, BlockNode)):
                new_child_nodes = self.get_next_button_nodes(node)
                nodes.extend(new_child_nodes)
            elif isinstance(node, InclusionNode) and node.filename == NEXT_BUTTON_TEMPLATE_PATH:
                nodes.append(node)
        return nodes

    def check_next_button(self):
        next_button_nodes = self.get_next_button_nodes(self.root)
        return len(next_button_nodes) > 0


def check_next_button(root):
    check = TemplateCheckNextButton(root)
    return check.check_next_button()


def has_valid_encoding(file_name):
    try:
        # need to open the file with an explicit encoding='utf8'
        # otherwise Windows may use another encoding if.
        # io.open provides the encoding= arg and is Py2/Py3 compatible
        with io.open(file_name, 'r', encoding='utf8') as f:
            template_string = f.read()
        force_text(template_string)
    except UnicodeDecodeError:
        return False
    return True


Line = namedtuple('Line', ('source', 'lineno', 'start', 'end'))


def format_error_line(line):
    # We need to make sure here that the output does not contain any unicode
    # characters. Django's check framework cannot print errors that contain
    # unicode.
    source = line.source
    source = unicodedata.normalize('NFKD', source).encode('ascii', 'replace')
    return '{line.lineno:4d} | {source}'.format(line=line, source=source)


def split_source_lines(source):
    """
    Split source string into a list of ``Line`` objects. They contain
    contextual information like line number, start position, end position.
    """
    lines = source.splitlines(True)
    start = 0
    annotated_lines = []
    for i, line in enumerate(lines):
        # Windows line endings end with '\r\n'.
        if line.endswith('\r\n'):
            ending_length = 2
        # In case of '\n' or '\r' ending the line.
        else:
            ending_length = 1
        end = start + len(line)
        annotated_lines.append(Line(
            # Don't include line endings in snippet source.
            source=line[:-ending_length],
            lineno=i + 1,
            start=start,
            end=end))
        start = end
    return annotated_lines


def format_source_snippet(source, arrow_position, context=5):
    """
    Display parts of a source file with an arrow pointing at an exact location.
    Will display ``context`` number of lines before and after the arrow
    position.

    Example::

          15 |     <p>
          16 |         Please provide your information in the form below.
          17 |     </p>
          18 |
          19 |     {% formrow form.my_field with label = "foo" %}
        -----------^
          20 |
          21 |     {% next_button %}
          22 | {% endblock %}
    """
    lines = split_source_lines(source)
    error_line = 0
    for line in lines:
        if line.start <= arrow_position < line.end:
            error_line = line
            break
    start_context = max(error_line.lineno - 1 - context, 0)
    end_context = min(error_line.lineno + context, len(lines))
    before = lines[start_context:error_line.lineno]
    after = lines[error_line.lineno:end_context]

    error_prefix = max(len(str(error_line.lineno)), 4) + len(' | ')
    error_length = max(arrow_position - error_line.start, 0)
    error_arrow = ('-' * (error_prefix + error_length)) + '^'
    return '\n'.join(chain(
        [format_error_line(line) for line in before],
        [error_arrow],
        [format_error_line(line) for line in after],
    ))



from typing import List
import logging
from collections import OrderedDict, defaultdict
from unittest import mock


from django.conf import settings
import pytest
import sys

import otree.session
import otree.common_internal

from .bot import ParticipantBot
import datetime
import os
import codecs
import otree.export
from otree.constants_internal import AUTO_NAME_BOTS_EXPORT_FOLDER
from otree.models_concrete import ParticipantToPlayerLookup
from otree.models import Session, Participant

logger = logging.getLogger(__name__)


class SessionBotRunner:
    def __init__(self, bots: List[ParticipantBot]):
        self.bots = OrderedDict()

        for bot in bots:
            self.bots[bot.participant_id] = bot

    def play(self):
        '''round-robin'''
        self.open_start_urls()
        loops_without_progress = 0
        while True:
            if len(self.bots) == 0:
                return
            # bots got stuck if there's 2 wait pages in a row
            if loops_without_progress > 10:
                raise AssertionError('Bots got stuck')
            # store in a separate list so we don't mutate the iterable
            playable_ids = list(self.bots.keys())
            progress_made = False
            for pk in playable_ids:
                bot = self.bots[pk]
                if bot.on_wait_page():
                    pass
                else:
                    try:
                        submission = next(bot.submits_generator)
                    except StopIteration:
                        # this bot is finished
                        self.bots.pop(pk)
                        progress_made = True
                    else:
                        bot.submit(submission)
                        progress_made = True
            if not progress_made:
                loops_without_progress += 1

    def open_start_urls(self):
        for bot in self.bots.values():
            bot.open_start_url()


def make_bots(*, session_pk, case_number, use_browser_bots) -> List[ParticipantBot]:
    update_kwargs = {'_is_bot': True}
    if use_browser_bots:
        update_kwargs['is_browser_bot'] = True

    Participant.objects.filter(session_id=session_pk).update(**update_kwargs)
    bots = []

    # can't use .distinct('player_pk') because it only works on Postgres
    # this implicitly orders by round also
    lookups = ParticipantToPlayerLookup.objects.filter(
        session_pk=session_pk).order_by('page_index')

    seen_players = set()
    lookups_per_participant = defaultdict(list)
    for lookup in lookups:
        player_identifier = (lookup.app_name, lookup.player_pk)
        if not player_identifier in seen_players:
            lookups_per_participant[lookup.participant_code].append(lookup)
            seen_players.add(player_identifier)

    for participant_code, lookups in lookups_per_participant.items():
        bot = ParticipantBot(lookups=lookups, case_number=case_number)
        bots.append(bot)

    return bots


def run_bots(session: Session, case_number=None):
    bot_list = make_bots(
        session_pk=session.pk, case_number=case_number, use_browser_bots=False
    )
    runner = SessionBotRunner(bots=bot_list)
    runner.play()


# function name needs to start with "test" for pytest to discover it
# in this module
@pytest.mark.django_db(transaction=True)
def test_all_bots_for_session_config(
        session_config_name, num_participants, *, run_export):
    config_name = session_config_name
    session_config = otree.session.SESSION_CONFIGS_DICT[config_name]

    # num_bots is deprecated, because the old default of 12 or 6 was too
    # much, and it doesn't make sense to
    if num_participants is None:
        num_participants = session_config['num_demo_participants']

    num_bot_cases = session_config.get_num_bot_cases()
    for case_number in range(num_bot_cases):
        if num_bot_cases > 1:
            logger.info("Creating '{}' session (test case {})".format(
                config_name, case_number))
        else:
            logger.info("Creating '{}' session".format(config_name))

        session = otree.session.create_session(
            session_config_name=config_name,
            num_participants=num_participants,
        )

        run_bots(session, case_number=case_number)
        logger.info('Bots completed session')
    if run_export:
        # bug: if the user tests multiple session configs,
        # the data will only be exported for the last session config.
        # this is because the DB is cleared after each test case.
        # fix later if this becomes high priority
        export_path = pytest.config.option.export_path

        now = datetime.datetime.now()

        if export_path == AUTO_NAME_BOTS_EXPORT_FOLDER:
            # oTree convention to prefix __temp all temp folders.
            export_path = now.strftime('__temp_bots_%b%d_%Hh%Mm%S.%f')[:-5] + 's'

        if os.path.isdir(export_path):
            msg = "Directory '{}' already exists".format(export_path)
            raise IOError(msg)

        os.makedirs(export_path)

        for app in settings.INSTALLED_OTREE_APPS:
            model_module = otree.common_internal.get_models_module(app)
            if model_module.Player.objects.exists():
                fname = "{}.csv".format(app)
                fpath = os.path.join(export_path, fname)
                with codecs.open(fpath, "w", encoding="utf8") as fp:
                    otree.export.export_app(app, fp, file_extension='csv')

        logger.info('Exported CSV to folder "{}"'.format(export_path))


def run_pytests(**kwargs):
    session_config_name = kwargs['session_config_name']
    num_participants = kwargs['num_participants']
    verbosity = kwargs['verbosity']

    this_module = sys.modules[__name__]

    # '-s' is to see print output
    # --tb=short is to show short tracebacks. I think this is
    # more expected and less verbose.
    # With the default pytest long tracebacks,
    # often the code that gets printed is in otree-core, which is not relevant.
    # also, this is better than using --tb=native, which loses line breaks
    # when a unicode char is contained in the output, and also doesn't get
    # color coded with colorama, the way short tracebacks do.
    argv = [
        this_module.__file__,
        '-s',
        '--tb', 'short'
    ]
    if verbosity == 0:
        argv.append('--quiet')
    if verbosity == 2:
        argv.append('--verbose')

    if session_config_name:
        argv.extend(['--session_config_name', session_config_name])
    if num_participants:
        argv.extend(['--num_participants', num_participants])
    if kwargs['preserve_data']:
        argv.append('--preserve_data')
    if kwargs['export_path']:
        argv.extend(['--export_path', kwargs['export_path']])

    # same hack as in resetdb code
    # because this method uses the serializer
    # it breaks if the app has migrations but they aren't up to date
    otree.common_internal.patch_migrations_module()
    return pytest.main(argv)



#!/usr/bin/env python
import os
import sys

from honcho.manager import Manager
from channels.log import setup_logger
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Run timeoutworker (plus botworker).'

    def add_arguments(self, parser):
        BaseCommand.add_arguments(self, parser)

    def get_env(self, options):
        return os.environ.copy()

    def handle(self, *args, **options):
        self.verbosity = options.get('verbosity', 1)
        self.logger = setup_logger('django.channels', self.verbosity)
        manager = self.get_honcho_manager(options)
        manager.loop()
        sys.exit(manager.returncode)

    def get_honcho_manager(self, options):

        manager = Manager()

        # if I change these, I need to modify the ServerCheck also
        manager.add_process(
            'botworker',
            'otree botworker',
            quiet=False,
            env=os.environ.copy()
        )
        manager.add_process(
            'timeoutworkeronly',
            'otree timeoutworkeronly',
            quiet=False,
            env=os.environ.copy()
        )

        return manager



import json
import logging
import django.db
import django.utils.timezone
import traceback
import time
from channels import Group
from channels.generic.websockets import JsonWebsocketConsumer
from django.core.signing import Signer, BadSignature
import otree.session
from otree.channels.utils import get_chat_group
from otree.models import Participant, Session
from otree.models_concrete import (
    CompletedGroupWaitPage, CompletedSubsessionWaitPage, ChatMessage)
from otree.common_internal import (
    get_models_module
)
import otree.channels.utils as channel_utils
from otree.models_concrete import (
    FailedSessionCreation, ParticipantRoomVisit,
    FAILURE_MESSAGE_MAX_LENGTH, BrowserBotsLauncherSessionCode)
from otree.room import ROOM_DICT
import otree.bots.browser
from otree.export import export_wide, export_app
import io
import base64
import datetime
from django.conf import settings

logger = logging.getLogger(__name__)


class InvalidWebSocketParams(Exception):
    '''exception to raise when websocket params are invalid'''

class OTreeJsonWebsocketConsumer(JsonWebsocketConsumer):
    '''
    THIS IS NOT PUBLIC API.
    Third party apps should not subclass this.
    Either copy this class into your code,
    or subclass directly from JsonWebsocketConsumer,
    '''

    def raw_connect(self, message, **kwargs):
        try:
            super().raw_connect(message, **kwargs)
        except InvalidWebSocketParams:
            logger.warning('Rejected request: {}'.format(self.path))

    def clean_kwargs(self, **kwargs):
        '''
        subclasses should override if the route receives a comma-separated params arg.
        otherwise, this just passes the route kwargs as is (usually there is just one).
        The output of this method is passed to self.group_name(), self.post_connect,
        and self.pre_disconnect, so within each class, all 3 of those methods must
        accept the same args (or at least take a **kwargs wildcard, if the args aren't used)
        '''
        return kwargs

    def group_name(self, **kwargs):
        raise NotImplementedError()

    def connection_groups(self, **kwargs):
        kwargs = self.clean_kwargs(**kwargs)
        group_name = self.group_name(**kwargs)
        return [group_name]

    def connect(self, message, **kwargs):
        # don't send accept: True until we upgrade to channels 1.0+
        # self.message.reply_channel.send({"accept": True})
        # only wrap this for connect. it's not as high-priority for
        # disconnect or receive.
        kwargs = self.clean_kwargs(**kwargs)
        self.post_connect(**kwargs)

    def post_connect(self, **kwargs):
        pass

    def disconnect(self, message, **kwargs):
        kwargs = self.clean_kwargs(**kwargs)
        self.pre_disconnect(**kwargs)

    def pre_disconnect(self, **kwargs):
        pass

    def receive(self, content, **kwargs):
        kwargs = self.clean_kwargs(**kwargs)
        self.post_receive(content, **kwargs)

    def post_receive(self, content, **kwargs):
        pass


class GroupByArrivalTime(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        session_pk, page_index, app_name, player_id = params.split(',')
        return {
            'app_name': app_name,
            'session_pk': int(session_pk),
            'page_index': int(page_index),
            'player_id': int(player_id)
        }

    def group_name(self, app_name, player_id, page_index, session_pk):
        gn = channel_utils.gbat_group_name(
            session_pk, page_index)
        return gn

    def post_connect(self, app_name, player_id, page_index, session_pk):
        models_module = get_models_module(app_name)
        group_id_in_subsession = models_module.Group.objects.filter(
            player__id=player_id).values_list(
            'id_in_subsession', flat=True)[0]

        ready = CompletedGroupWaitPage.objects.filter(
            page_index=page_index,
            id_in_subsession=int(group_id_in_subsession),
            session_id=session_pk,
            ).exists()
        if ready:
            self.send({'status': 'ready'})


class WaitPage(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        session_pk, page_index, group_id_in_subsession = params.split(',')
        return {
            'session_pk': int(session_pk),
            'page_index': int(page_index),
            # don't convert group_id_in_subsession to int yet, it might be null
            'group_id_in_subsession': group_id_in_subsession,
        }

    def group_name(self, session_pk, page_index, group_id_in_subsession):
        return channel_utils.wait_page_group_name(
                session_pk, page_index, group_id_in_subsession)

    def post_connect(self, session_pk, page_index, group_id_in_subsession):
        # in case message was sent before this web socket connects
        if group_id_in_subsession:
            ready = CompletedGroupWaitPage.objects.filter(
                page_index=page_index,
                id_in_subsession=int(group_id_in_subsession),
                session_id=session_pk,
                ).exists()
        else:  # subsession
            ready = CompletedSubsessionWaitPage.objects.filter(
                page_index=page_index,
                session_id=session_pk,
                ).exists()
        if ready:
            self.send({'status': 'ready'})


class AutoAdvance(OTreeJsonWebsocketConsumer):
    def clean_kwargs(self, params):
        participant_code, page_index = params.split(',')
        return {
            'participant_code': participant_code,
            'page_index': int(page_index),
        }

    def group_name(self, page_index, participant_code):
        return 'auto-advance-{}'.format(participant_code)

    def post_connect(self, page_index, participant_code):
        # in case message was sent before this web socket connects
        result = Participant.objects.filter(
                code=participant_code).values_list(
            '_index_in_pages', flat=True)
        try:
            page_should_be_on = result[0]
        except IndexError:
            # doesn't get shown because not yet localized
            self.send({'error': 'Participant not found in database.'})
            return
        if page_should_be_on > page_index:
            self.send({'auto_advanced': True})


def create_session(message):
    group = Group(message['channels_group_name'])

    kwargs = message['kwargs']

    try:
        session = otree.session.create_session(**kwargs)
        if message['use_browser_bots']:
            otree.bots.browser.initialize_session(
                session_pk=session.pk,
                case_number=None
            )
        session.ready_for_browser = True
        session.save()
    except Exception as e:

        # full error message is printed to console (though sometimes not?)
        error_message = 'Failed to create session: "{}"'.format(e)
        traceback_str = traceback.format_exc()
        group.send(
            {'text': json.dumps(
                {
                    'error': error_message,
                    'traceback': traceback_str,
                })}
        )
        FailedSessionCreation.objects.create(
            pre_create_id=kwargs['pre_create_id'],
            message=error_message[:FAILURE_MESSAGE_MAX_LENGTH],
            traceback=traceback_str
        )
        raise

    group.send(
        {'text': json.dumps(
            {'status': 'ready'})}
    )

    if 'room_name' in kwargs:
        Group(channel_utils.room_participants_group_name(kwargs['room_name'])).send(
            {'text': json.dumps(
                {'status': 'session_ready'})}
        )


class WaitForSession(OTreeJsonWebsocketConsumer):
    def clean_kwargs(self, **kwargs):
        return kwargs

    def group_name(self, pre_create_id):
        return channel_utils.create_session_group_name(pre_create_id)

    def post_connect(self, pre_create_id):

        group_name = self.group_name(pre_create_id)

        # in case message was sent before this web socket connects
        if Session.objects.filter(
                _pre_create_id=pre_create_id, ready_for_browser=True).exists():
            self.group_send(group_name, {'status': 'ready'})
        else:
            failure = FailedSessionCreation.objects.filter(
                pre_create_id=pre_create_id
            ).first()
            if failure:
                self.group_send(group_name,
                        {'error': failure.message,
                         'traceback': failure.traceback})


class RoomAdmin(OTreeJsonWebsocketConsumer):

    def group_name(self, room):
        return 'room-admin-{}'.format(room)

    def post_connect(self, room):

        room_object = ROOM_DICT[room]

        now = time.time()
        stale_threshold = now - 15
        present_list = ParticipantRoomVisit.objects.filter(
            room_name=room_object.name,
            last_updated__gte=stale_threshold,
        ).values_list('participant_label', flat=True)

        # make it JSON serializable
        present_list = list(present_list)

        self.send({
            'status': 'load_participant_lists',
            'participants_present': present_list,
        })

        # prune very old visits -- don't want a resource leak
        # because sometimes not getting deleted on WebSocket disconnect
        very_stale_threshold = now - 10*60
        ParticipantRoomVisit.objects.filter(
            room_name=room_object.name,
            last_updated__lt=very_stale_threshold,
        ).delete()


class RoomParticipant(OTreeJsonWebsocketConsumer):

    def clean_kwargs(self, params):
        room_name, participant_label, tab_unique_id = params.split(',')
        return {
            'room_name': room_name,
            'participant_label': participant_label,
            'tab_unique_id': tab_unique_id,
        }

    def group_name(self, room_name, participant_label, tab_unique_id):
        return channel_utils.room_participants_group_name(room_name)

    def post_connect(self, room_name, participant_label, tab_unique_id):
        if room_name in ROOM_DICT:
            room = ROOM_DICT[room_name]
        else:
            # doesn't get shown because not yet localized
            self.send({'error': 'Invalid room name "{}".'.format(room_name)})
            return
        if room.has_session():
            self.send({'status': 'session_ready'})
        else:
            try:
                ParticipantRoomVisit.objects.create(
                    participant_label=participant_label,
                    room_name=room_name,
                    tab_unique_id=tab_unique_id,
                    last_updated=time.time(),
                )
            except django.db.IntegrityError:
                # possible that the tab connected twice
                # without disconnecting in between
                # because of WebSocket failure
                # tab_unique_id is unique=True,
                # so this will throw an integrity error.
                # 2017-09-17: I saw the integrityerror on macOS.
                # previously, we logged this, but i see no need to do that.
                pass
            self.group_send(
                'room-admin-{}'.format(room_name),
                {
                    'status': 'add_participant',
                    'participant': participant_label
                }
            )

    def pre_disconnect(self, room_name, participant_label, tab_unique_id):

        if room_name in ROOM_DICT:
            room = ROOM_DICT[room_name]
        else:
            # doesn't get shown because not yet localized
            self.send({'error': 'Invalid room name "{}".'.format(room_name)})
            return

        # should use filter instead of get,
        # because if the DB is recreated,
        # the record could already be deleted
        ParticipantRoomVisit.objects.filter(
            participant_label=participant_label,
            room_name=room_name,
            tab_unique_id=tab_unique_id).delete()

        if room.has_participant_labels():
            if not ParticipantRoomVisit.objects.filter(
                participant_label=participant_label,
                room_name=room_name
            ).exists():
                # it's ok if there is a race condition --
                # in JS removing a participant is idempotent
                self.group_send(
                    'room-admin-{}'.format(room_name),
                    {
                        'status': 'remove_participant',
                        'participant': participant_label
                    }
                )
        else:
            self.group_send(
                'room-admin-{}'.format(room_name),
                {
                    'status': 'remove_participant',
                }
            )


class BrowserBotsLauncher(OTreeJsonWebsocketConsumer):

    def group_name(self, session_code):
        return channel_utils.browser_bots_launcher_group(session_code)


class BrowserBot(OTreeJsonWebsocketConsumer):

    def group_name(self):
        return 'browser_bot_wait'

    def post_connect(self):
        launcher_session_info = BrowserBotsLauncherSessionCode.objects.first()
        if launcher_session_info:
            self.send({'status': 'session_ready'})


class ChatConsumer(OTreeJsonWebsocketConsumer):

    # Set to True if you want it, else leave it out
    strict_ordering = False

    def clean_kwargs(self, params):

        signer = Signer(sep='/')
        try:
            original_params = signer.unsign(params)
        except BadSignature:
            raise InvalidWebSocketParams

        channel, participant_id = original_params.split('/')

        return {
            'channel': channel,
            'participant_id': int(participant_id),
        }

    def group_name(self, channel, participant_id):
        return get_chat_group(channel)

    def post_connect(self, **kwargs):

        history = ChatMessage.objects.filter(
            channel=kwargs['channel']).order_by('timestamp').values(
                'nickname', 'body', 'participant_id'
        )

        # Convert ValuesQuerySet to list
        self.send(list(history))

    def post_receive(self, content, channel, participant_id):
        content['channel'] = channel
        content['participant_id'] = participant_id

        # in the Channels docs, the example has a separate msg_consumer
        # channel, so this can be done asynchronously.
        # but i think the perf is probably good enough.
        # moving into here for simplicity, especially for testing.
        nickname_signed = content['nickname_signed']
        nickname = Signer().unsign(nickname_signed)
        channel = content['channel']
        channels_group = get_chat_group(channel)

        body = content['body']
        participant_id = content['participant_id']

        chat_message = {
            'nickname': nickname,
            'body': body,
            'participant_id': participant_id
        }

        Group(channels_group).send({'text': json.dumps([chat_message])})

        ChatMessage.objects.create(
            participant_id=participant_id,
            channel=channel,
            body=body,
            nickname=nickname
        )


class ExportData(OTreeJsonWebsocketConsumer):
    # access to self.message.user for auth
    http_user = True

    def post_receive(self, content: dict):
        '''
        if an app name is given, export the app.
        otherwise, export all the data (wide).
        don't need time_spent or chat yet, they are quick enough
        '''

        # authenticate
        # maybe it should be is_superuser or something else more specific
        # but this is to be consistent with the rest of Django's login
        if settings.AUTH_LEVEL and not self.message.user.is_authenticated:
            logger.warning(
                'rejected access to data export through non-authenticated '
                'websocket'
            )
            return

        file_extension = content['file_extension']
        app_name = content.get('app_name')

        if file_extension == 'xlsx':
            mime_type = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
            IOClass = io.BytesIO
        else:
            mime_type = 'text/csv'
            IOClass = io.StringIO

        iso_date = datetime.date.today().isoformat()
        with IOClass() as fp:
            if app_name:
                export_app(app_name, fp, file_extension=file_extension)
                file_name_prefix = app_name
            else:
                export_wide(fp, file_extension=file_extension)
                file_name_prefix = 'all_apps_wide'
            data = fp.getvalue()

        file_name = '{}_{}.{}'.format(
            file_name_prefix, iso_date, file_extension)

        if file_extension == 'xlsx':
            data = base64.b64encode(data).decode('utf-8')

        content.update({
            'file_name': file_name,
            'data': data,
            'mime_type': mime_type,
        })
        self.send(content)

    def connection_groups(self, **kwargs):
        return []



import codecs
from collections import OrderedDict

import schema

from otree.models_concrete import RoomToSession
from otree.common_internal import (
    add_params_to_url, make_hash, validate_alphanumeric)
from django.conf import settings
from django.urls import reverse
from django.db import transaction

class Room(object):

    def __init__(self, config_dict):
        self.participant_label_file = config_dict.get('participant_label_file')

        self.name = validate_alphanumeric(
            config_dict['name'],
            identifier_description='settings.ROOMS room name')
        self.display_name = config_dict['display_name']
        # secure URLs are complicated, don't use them by default
        self.use_secure_urls = config_dict['use_secure_urls']
        if self.use_secure_urls and not self.participant_label_file:
            raise ValueError(
                'Room "{}": you must either set "participant_label_file", '
                'or set "use_secure_urls": False'.format(self.name)
            )

    def has_session(self):
        return self.get_session() is not None

    def get_session(self):
        try:
            return RoomToSession.objects.select_related('session').get(
                room_name=self.name).session
        except RoomToSession.DoesNotExist:
            return None

    def set_session(self, session):
        with transaction.atomic():
            RoomToSession.objects.filter(room_name=self.name).delete()
            if session:
                RoomToSession.objects.create(
                    room_name=self.name,
                    session=session
                )

    def has_participant_labels(self):
        return bool(self.participant_label_file)

    def get_participant_labels(self):
        '''
        Decided to just re-read the file on every request,
        rather than loading in the DB. Reasons:

        (1) Simplifies the code; we don't need an ExpectedRoomParticipant model,
            and don't need to load data into there (which involves considerations
            of race conditions)
        (2) Don't need any complicated rule deciding when to reload the file,
            whether it's upon starting the process or resetting the database,
            or both. Should the status be stored in the DB or in the process?
        (3) Checking if a given participant label is in the file is actually faster
            than looking it up in the DB table, even with .filter() and and index!
            (tested on Postgres with 10000 iterations: 17s vs 18s)
        '''
        encodings = ['ascii', 'utf-8', 'utf-16']
        for e in encodings:
            try:
                plabel_path = self.participant_label_file
                with codecs.open(plabel_path, "r", encoding=e) as f:
                    seen = set()
                    labels = []
                    for line in f:
                        label = line.strip()
                        if not label:
                            continue
                        label = validate_alphanumeric(
                            line.strip(),
                            identifier_description='participant label'
                        )
                        if label not in seen:
                            labels.append(label)
                            seen.add(label)
            except UnicodeDecodeError:
                continue
            except FileNotFoundError:
                msg = (
                    'settings.ROOMS: The room "{}" references '
                    ' nonexistent participant_label_file "{}".'
                )
                raise FileNotFoundError(
                    msg.format(self.name, self.participant_label_file)
                ) from None
            else:
                return labels
        raise Exception(
            'settings.ROOMS: participant_label_file "{}" '
            'not encoded correctly.'.format(self.participant_label_file)
        )

    def get_room_wide_url(self, request):
        url = reverse('AssignVisitorToRoom', args=(self.name,))
        return request.build_absolute_uri(url)

    def get_participant_urls(self, request):
        participant_urls = []
        room_base_url = reverse('AssignVisitorToRoom', args=(self.name,))
        room_base_url = request.build_absolute_uri(room_base_url)

        if self.has_participant_labels():
            for label in self.get_participant_labels():
                params = {'participant_label': label}
                if self.use_secure_urls:
                    params['hash'] = make_hash(label)
                participant_url = add_params_to_url(room_base_url, params)
                participant_urls.append(participant_url)

        return participant_urls


def augment_room(room, ROOM_DEFAULTS):
    new_room = {'doc': ''}
    new_room.update(ROOM_DEFAULTS)
    new_room.update(room)
    return new_room


def get_room_dict():
    room_defaults_schema = schema.Schema(
        {
            schema.Optional('use_secure_urls', default=False): bool,
            schema.Optional('participant_label_file'): str,
            schema.Optional('doc'): str,
        }
    )

    room_schema = schema.Schema(
        {
            'name': str,
            'display_name': str,
            schema.Optional('use_secure_urls'): bool,
            schema.Optional('participant_label_file'): str,
            schema.Optional('doc'): str,
        }
    )

    ROOM_DICT = OrderedDict()
    ROOM_DEFAULTS = getattr(settings, 'ROOM_DEFAULTS', {})
    try:
        ROOM_DEFAULTS = room_defaults_schema.validate(ROOM_DEFAULTS)
    except schema.SchemaError as e:
        raise (ValueError('settings.ROOM_DEFAULTS: {}'.format(e))) from None
    for room in getattr(settings, 'ROOMS', []):
        room = augment_room(room, ROOM_DEFAULTS)
        try:
            room = room_schema.validate(room)
        except schema.SchemaError as e:
            raise(ValueError('settings.ROOMS: {}'.format(e))) from None
        room_object = Room(room)
        ROOM_DICT[room_object.name] = room_object
    return ROOM_DICT

ROOM_DICT = get_room_dict()



from django.template.base import FilterExpression, Node
from django.template.smartif import TokenBase

'''
Make templates strict.
This has the potential to break existing templates, especially those in Django core.
But I tested by dropping Django's widget_tests folder into our test suite and 
running those tests, and did not get any regressions (same number of tests fail,
and don't get any InvalidVariableError's).
Other parts of Django with templates, such as admin, admindocs, postgres(?), GIS,
are not used by oTree. So we just need to consider users' templates,
which are unlikely to rely on silent failures.
'''

def patch_filter_expression():
    '''
    don't allow code like {{ bogus }} or {{ player.bogus }} to fail silently
    '''

    # can't fix this through inheritance because the class name
    # is hardcoded in django.template.base.Parser.compile_filter()

    original_resolve = FilterExpression.resolve
    # block {{ bogus }} but allow {{ bogus|default:1 }}, which seems to be
    # a common pattern for defaulting if a var is not found
    def resolve(self, context, ignore_failures=False):
        if self.filters:
            # allow {{ bogus|default:1 }}
            return original_resolve(self, context, ignore_failures=True)
        else:
            # this covers the following cases:
            # {{ bogus }}
            # {% if bogus %}
            # if-statements use ignore_failures=True,
            # because of TemplateLiteral.eval()
            # a bit risky to override, because
            # if-statements on nonexistent variables exist in django core,
            # such as in django/forms/templates/django/forms/widgets:
            # {% if widget.attrs.class %}
            return original_resolve(self, context, ignore_failures=False)
    FilterExpression.resolve = resolve

def patch_smartif():
    '''
    SmartIf is for if-statements with multiple tokens, like {% if bogus == 1 %}
    == is by far the most commonly used
    but we need to patch all operators, because now when FilterExpression.resolve()
    can raise, these functions will swallow the exception and return False,
    which can silently change the behavior of certain template expressions,
    which is unacceptable.
    For example, if we don't do this, then
    {% bogus != 'ok' %}
    will evaluate to False!
    '''

    def make_infix_eval(func):
        '''see infix()'s Operator.eval()'''
        def new_eval(self, context):
            return func(context, self.first, self.second)
        return new_eval

    def make_prefix_eval(func):
        '''see prefix()'s Operator.eval()'''
        def new_eval(self, context):
            return func(context, self.first)
        return new_eval

    infix_operators = {
        'or': lambda context, x, y: x.eval(context) or y.eval(context),
        'and': lambda context, x, y: x.eval(context) and y.eval(context),
        'in': lambda context, x, y: x.eval(context) in y.eval(context),
        'not in': lambda context, x, y: x.eval(context) not in y.eval(context),
        'is': lambda context, x, y: x.eval(context) is y.eval(context),
        'is not': lambda context, x, y: x.eval(context) is not y.eval(context),
        '==': lambda context, x, y: x.eval(context) == y.eval(context),
        '!=': lambda context, x, y: x.eval(context) != y.eval(context),
        '>': lambda context, x, y: x.eval(context) > y.eval(context),
        '>=': lambda context, x, y: x.eval(context) >= y.eval(context),
        '<': lambda context, x, y: x.eval(context) < y.eval(context),
        '<=': lambda context, x, y: x.eval(context) <= y.eval(context),
    }

    from django.template.smartif import OPERATORS
    for operator, func in infix_operators.items():
        OPERATORS[operator].eval = make_infix_eval(func)
    OPERATORS['not'].eval = make_prefix_eval(lambda context, x: not x.eval(context))


def patch_28935():
    # if we do the above patch, we also need a patch for the open issue:
    # https://code.djangoproject.com/ticket/28935
    # usually in oTree the problem will be in the user's (child) template
    def render_annotated(self, context):
        try:
            return self.render(context)
        except Exception as e:
            if context.template.engine.debug and not hasattr(e, 'template_debug'):
                e.template_debug = context.template.get_exception_info(e, self.token)
            raise
    Node.render_annotated = render_annotated


def patch_template_silent_failures():
    patch_filter_expression()
    patch_smartif()
    patch_28935()



